"""Router for MCP interface providing Model Context Protocol endpoints."""

import logging
from typing import TYPE_CHECKING, List, Optional, cast
from uuid import uuid4

from fastmcp import FastMCP
from fastmcp.server.http import (
    StarletteWithLifespan,
)

from agno.db.base import AsyncBaseDb, SessionType
from agno.db.schemas import UserMemory
from agno.os.routers.memory.schemas import (
    UserMemorySchema,
)
from agno.os.schema import (
    AgentSummaryResponse,
    ConfigResponse,
    InterfaceResponse,
    SessionSchema,
    TeamSummaryResponse,
    WorkflowSummaryResponse,
)
from agno.os.utils import (
    get_agent_by_id,
    get_db,
    get_team_by_id,
    get_workflow_by_id,
)
from agno.run.agent import RunOutput
from agno.run.team import TeamRunOutput
from agno.run.workflow import WorkflowRunOutput

if TYPE_CHECKING:
    from agno.os.app import AgentOS

logger = logging.getLogger(__name__)


def get_mcp_server(
    os: "AgentOS",
) -> StarletteWithLifespan:
    """Attach MCP routes to the provided router."""

    # Create an MCP server
    mcp = FastMCP(os.name or "AgentOS")

    @mcp.tool(
        name="get_agentos_config",
        description="Get the configuration of the AgentOS",
        tags={"core"},
        output_schema=ConfigResponse.model_json_schema(),
    )  # type: ignore
    async def config() -> ConfigResponse:
        return ConfigResponse(
            os_id=os.id or "AgentOS",
            description=os.description,
            available_models=os.config.available_models if os.config else [],
            databases=[db.id for db in os.dbs.values()],
            chat=os.config.chat if os.config else None,
            session=os._get_session_config(),
            memory=os._get_memory_config(),
            knowledge=os._get_knowledge_config(),
            evals=os._get_evals_config(),
            metrics=os._get_metrics_config(),
            agents=[AgentSummaryResponse.from_agent(agent) for agent in os.agents] if os.agents else [],
            teams=[TeamSummaryResponse.from_team(team) for team in os.teams] if os.teams else [],
            workflows=[WorkflowSummaryResponse.from_workflow(w) for w in os.workflows] if os.workflows else [],
            interfaces=[
                InterfaceResponse(type=interface.type, version=interface.version, route=interface.prefix)
                for interface in os.interfaces
            ],
        )

    @mcp.tool(name="run_agent", description="Run an agent", tags={"core"})  # type: ignore
    async def run_agent(agent_id: str, message: str) -> RunOutput:
        agent = get_agent_by_id(agent_id, os.agents)
        if agent is None:
            raise Exception(f"Agent {agent_id} not found")
        return await agent.arun(message)

    @mcp.tool(name="run_team", description="Run a team", tags={"core"})  # type: ignore
    async def run_team(team_id: str, message: str) -> TeamRunOutput:
        team = get_team_by_id(team_id, os.teams)
        if team is None:
            raise Exception(f"Team {team_id} not found")
        return await team.arun(message)

    @mcp.tool(name="run_workflow", description="Run a workflow", tags={"core"})  # type: ignore
    async def run_workflow(workflow_id: str, message: str) -> WorkflowRunOutput:
        workflow = get_workflow_by_id(workflow_id, os.workflows)
        if workflow is None:
            raise Exception(f"Workflow {workflow_id} not found")
        return await workflow.arun(message)

    # Session Management Tools
    @mcp.tool(name="get_sessions_for_agent", description="Get list of sessions for an agent", tags={"session"})  # type: ignore
    async def get_sessions_for_agent(
        agent_id: str,
        db_id: str,
        user_id: Optional[str] = None,
        sort_by: str = "created_at",
        sort_order: str = "desc",
    ):
        db = get_db(os.dbs, db_id)
        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            sessions = await db.get_sessions(
                session_type=SessionType.AGENT,
                component_id=agent_id,
                user_id=user_id,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )
        else:
            sessions = db.get_sessions(
                session_type=SessionType.AGENT,
                component_id=agent_id,
                user_id=user_id,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )

        return {
            "data": [SessionSchema.from_dict(session) for session in sessions],  # type: ignore
        }

    @mcp.tool(name="get_sessions_for_team", description="Get list of sessions for a team", tags={"session"})  # type: ignore
    async def get_sessions_for_team(
        team_id: str,
        db_id: str,
        user_id: Optional[str] = None,
        sort_by: str = "created_at",
        sort_order: str = "desc",
    ):
        db = get_db(os.dbs, db_id)
        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            sessions = await db.get_sessions(
                session_type=SessionType.TEAM,
                component_id=team_id,
                user_id=user_id,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )
        else:
            sessions = db.get_sessions(
                session_type=SessionType.TEAM,
                component_id=team_id,
                user_id=user_id,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )

        return {
            "data": [SessionSchema.from_dict(session) for session in sessions],  # type: ignore
        }

    @mcp.tool(name="get_sessions_for_workflow", description="Get list of sessions for a workflow", tags={"session"})  # type: ignore
    async def get_sessions_for_workflow(
        workflow_id: str,
        db_id: str,
        user_id: Optional[str] = None,
        sort_by: str = "created_at",
        sort_order: str = "desc",
    ):
        db = get_db(os.dbs, db_id)
        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            sessions = await db.get_sessions(
                session_type=SessionType.WORKFLOW,
                component_id=workflow_id,
                user_id=user_id,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )
        else:
            sessions = db.get_sessions(
                session_type=SessionType.WORKFLOW,
                component_id=workflow_id,
                user_id=user_id,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )

        return {
            "data": [SessionSchema.from_dict(session) for session in sessions],  # type: ignore
        }

    # Memory Management Tools
    @mcp.tool(name="create_memory", description="Create a new user memory", tags={"memory"})  # type: ignore
    async def create_memory(
        db_id: str,
        memory: str,
        user_id: str,
        topics: Optional[List[str]] = None,
    ) -> UserMemorySchema:
        db = get_db(os.dbs, db_id)
        user_memory = db.upsert_user_memory(
            memory=UserMemory(
                memory_id=str(uuid4()),
                memory=memory,
                topics=topics or [],
                user_id=user_id,
            ),
            deserialize=False,
        )
        if not user_memory:
            raise Exception("Failed to create memory")

        return UserMemorySchema.from_dict(user_memory)  # type: ignore

    @mcp.tool(name="get_memories_for_user", description="Get a list of memories for a user", tags={"memory"})  # type: ignore
    async def get_memories_for_user(
        user_id: str,
        sort_by: str = "updated_at",
        sort_order: str = "desc",
        db_id: Optional[str] = None,
    ):
        db = get_db(os.dbs, db_id)
        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            user_memories = await db.get_user_memories(
                user_id=user_id,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )
        else:
            user_memories = db.get_user_memories(
                user_id=user_id,
                sort_by=sort_by,
                sort_order=sort_order,
                deserialize=False,
            )
        return {
            "data": [UserMemorySchema.from_dict(user_memory) for user_memory in user_memories],  # type: ignore
        }

    @mcp.tool(name="update_memory", description="Update a memory", tags={"memory"})  # type: ignore
    async def update_memory(
        db_id: str,
        memory_id: str,
        memory: str,
        user_id: str,
    ) -> UserMemorySchema:
        db = get_db(os.dbs, db_id)
        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            user_memory = await db.upsert_user_memory(
                memory=UserMemory(
                    memory_id=memory_id,
                    memory=memory,
                    user_id=user_id,
                ),
                deserialize=False,
            )
        else:
            user_memory = db.upsert_user_memory(
                memory=UserMemory(
                    memory_id=memory_id,
                    memory=memory,
                    user_id=user_id,
                ),
                deserialize=False,
            )
        if not user_memory:
            raise Exception("Failed to update memory")

        return UserMemorySchema.from_dict(user_memory)  # type: ignore

    @mcp.tool(name="delete_memory", description="Delete a memory by ID", tags={"memory"})  # type: ignore
    async def delete_memory(
        db_id: str,
        memory_id: str,
    ) -> None:
        db = get_db(os.dbs, db_id)
        if isinstance(db, AsyncBaseDb):
            db = cast(AsyncBaseDb, db)
            await db.delete_user_memory(memory_id=memory_id)
        else:
            db.delete_user_memory(memory_id=memory_id)

    mcp_app = mcp.http_app(path="/mcp")
    return mcp_app



from dataclasses import dataclass


@dataclass
class ApiRoutes:
    """API routes for telemetry recordings"""

    # Runs
    RUN_CREATE: str = "/telemetry/runs"
    EVAL_RUN_CREATE: str = "/telemetry/evals"

    # OS launch
    AGENT_OS_LAUNCH: str = "/telemetry/os"



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Any, Dict, List, Optional


@dataclass
class UserMemory:
    """Model for User Memories"""

    memory: str
    memory_id: Optional[str] = None
    topics: Optional[List[str]] = None
    user_id: Optional[str] = None
    input: Optional[str] = None
    updated_at: Optional[datetime] = None
    feedback: Optional[str] = None

    agent_id: Optional[str] = None
    team_id: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        _dict = {
            "memory_id": self.memory_id,
            "memory": self.memory,
            "topics": self.topics,
            "updated_at": self.updated_at.isoformat() if self.updated_at else None,
            "input": self.input,
            "user_id": self.user_id,
            "agent_id": self.agent_id,
            "team_id": self.team_id,
            "feedback": self.feedback,
        }
        return {k: v for k, v in _dict.items() if v is not None}

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "UserMemory":
        data = dict(data)

        # Convert updated_at to datetime
        if updated_at := data.get("updated_at"):
            if isinstance(updated_at, (int, float)):
                data["updated_at"] = datetime.fromtimestamp(updated_at, tz=timezone.utc)
            else:
                data["updated_at"] = datetime.fromisoformat(updated_at)

        return cls(**data)



from abc import ABC, abstractmethod
from typing import Union

from agno.run.agent import RunInput
from agno.run.team import TeamRunInput


class BaseGuardrail(ABC):
    """Abstract base class for all guardrail implementations."""

    @abstractmethod
    def check(self, run_input: Union[RunInput, TeamRunInput]) -> None:
        """Perform synchronous guardrail check."""
        pass

    @abstractmethod
    async def async_check(self, run_input: Union[RunInput, TeamRunInput]) -> None:
        """Perform asynchronous guardrail check."""
        pass



from datetime import date, datetime, timedelta, timezone
from textwrap import dedent
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from agno.db.base import BaseDb, SessionType
from agno.db.postgres.utils import (
    get_dates_to_calculate_metrics_for,
)
from agno.db.schemas import UserMemory
from agno.db.schemas.evals import EvalFilterType, EvalRunRecord, EvalType
from agno.db.schemas.knowledge import KnowledgeRow
from agno.db.surrealdb import utils
from agno.db.surrealdb.metrics import (
    bulk_upsert_metrics,
    calculate_date_metrics,
    fetch_all_sessions_data,
    get_all_sessions_for_metrics_calculation,
    get_metrics_calculation_starting_date,
)
from agno.db.surrealdb.models import (
    TableType,
    deserialize_eval_run_record,
    deserialize_knowledge_row,
    deserialize_session,
    deserialize_sessions,
    deserialize_user_memories,
    deserialize_user_memory,
    desurrealize_eval_run_record,
    desurrealize_session,
    desurrealize_user_memory,
    get_schema,
    get_session_type,
    serialize_eval_run_record,
    serialize_knowledge_row,
    serialize_session,
    serialize_user_memory,
)
from agno.db.surrealdb.queries import COUNT_QUERY, WhereClause, order_limit_start
from agno.db.surrealdb.utils import build_client
from agno.session import Session
from agno.utils.log import log_debug, log_error, log_info
from agno.utils.string import generate_id

try:
    from surrealdb import BlockingHttpSurrealConnection, BlockingWsSurrealConnection, RecordID
except ImportError:
    raise ImportError("The `surrealdb` package is not installed. Please install it via `pip install surrealdb`.")


class SurrealDb(BaseDb):
    def __init__(
        self,
        client: Optional[Union[BlockingWsSurrealConnection, BlockingHttpSurrealConnection]],
        db_url: str,
        db_creds: dict[str, str],
        db_ns: str,
        db_db: str,
        session_table: Optional[str] = None,
        memory_table: Optional[str] = None,
        metrics_table: Optional[str] = None,
        eval_table: Optional[str] = None,
        knowledge_table: Optional[str] = None,
        id: Optional[str] = None,
    ):
        """
        Interface for interacting with a SurrealDB database.

        Args:
            client: A blocking connection, either HTTP or WS
        """
        if id is None:
            base_seed = db_url
            seed = f"{base_seed}#{db_db}"
            id = generate_id(seed)

        super().__init__(
            id=id,
            session_table=session_table,
            memory_table=memory_table,
            metrics_table=metrics_table,
            eval_table=eval_table,
            knowledge_table=knowledge_table,
        )
        self._client = client
        self._db_url = db_url
        self._db_creds = db_creds
        self._db_ns = db_ns
        self._db_db = db_db
        self._users_table_name: str = "agno_users"
        self._agents_table_name: str = "agno_agents"
        self._teams_table_name: str = "agno_teams"
        self._workflows_table_name: str = "agno_workflows"

    @property
    def client(self) -> Union[BlockingWsSurrealConnection, BlockingHttpSurrealConnection]:
        if self._client is None:
            self._client = build_client(self._db_url, self._db_creds, self._db_ns, self._db_db)
        return self._client

    @property
    def table_names(self) -> dict[TableType, str]:
        return {
            "agents": self._agents_table_name,
            "evals": self.eval_table_name,
            "knowledge": self.knowledge_table_name,
            "memories": self.memory_table_name,
            "sessions": self.session_table_name,
            "teams": self._teams_table_name,
            "users": self._users_table_name,
            "workflows": self._workflows_table_name,
        }

    def _table_exists(self, table_name: str) -> bool:
        response = self._query_one("INFO FOR DB", {}, dict)
        if response is None:
            raise Exception("Failed to retrieve database information")
        return table_name in response.get("tables", [])

    def _create_table(self, table_type: TableType, table_name: str):
        query = get_schema(table_type, table_name)
        self.client.query(query)

    def _get_table(self, table_type: TableType, create_table_if_not_found: bool = True):
        if table_type == "sessions":
            table_name = self.session_table_name
        elif table_type == "memories":
            table_name = self.memory_table_name
        elif table_type == "knowledge":
            table_name = self.knowledge_table_name
        elif table_type == "users":
            table_name = self._users_table_name
        elif table_type == "agents":
            table_name = self._agents_table_name
        elif table_type == "teams":
            table_name = self._teams_table_name
        elif table_type == "workflows":
            table_name = self._workflows_table_name
        elif table_type == "evals":
            table_name = self.eval_table_name
        elif table_type == "metrics":
            table_name = self.metrics_table_name
        else:
            raise NotImplementedError(f"Unknown table type: {table_type}")

        if create_table_if_not_found and not self._table_exists(table_name):
            self._create_table(table_type, table_name)

        return table_name

    def _query(
        self,
        query: str,
        vars: dict[str, Any],
        record_type: type[utils.RecordType],
    ) -> Sequence[utils.RecordType]:
        return utils.query(self.client, query, vars, record_type)

    def _query_one(
        self,
        query: str,
        vars: dict[str, Any],
        record_type: type[utils.RecordType],
    ) -> Optional[utils.RecordType]:
        return utils.query_one(self.client, query, vars, record_type)

    def _count(self, table: str, where_clause: str, where_vars: dict[str, Any], group_by: Optional[str] = None) -> int:
        total_count_query = COUNT_QUERY.format(
            table=table,
            where_clause=where_clause,
            group_clause="GROUP ALL" if group_by is None else f"GROUP BY {group_by}",
            group_fields="" if group_by is None else f", {group_by}",
        )
        count_result = self._query_one(total_count_query, where_vars, dict)
        total_count = count_result.get("count") if count_result else 0
        assert isinstance(total_count, int), f"Expected int, got {type(total_count)}"
        total_count = int(total_count)
        return total_count

    # --- Sessions ---
    def clear_sessions(self) -> None:
        """Delete all session rows from the database.

        Raises:
            Exception: If an error occurs during deletion.
        """
        table = self._get_table("sessions")
        _ = self.client.delete(table)

    def delete_session(self, session_id: str) -> bool:
        table = self._get_table(table_type="sessions")
        if table is None:
            return False
        res = self.client.delete(RecordID(table, session_id))
        return bool(res)

    def delete_sessions(self, session_ids: list[str]) -> None:
        table = self._get_table(table_type="sessions")
        if table is None:
            return

        records = [RecordID(table, id) for id in session_ids]
        self.client.query(f"DELETE FROM {table} WHERE id IN $records", {"records": records})

    def get_session(
        self,
        session_id: str,
        session_type: SessionType,
        user_id: Optional[str] = None,
        deserialize: Optional[bool] = True,
    ) -> Optional[Union[Session, Dict[str, Any]]]:
        r"""
        Read a session from the database.

        Args:
            session_id (str): ID of the session to read.
            session_type (SessionType): Type of session to get.
            user_id (Optional[str]): User ID to filter by. Defaults to None.
            deserialize (Optional[bool]): Whether to serialize the session. Defaults to True.

        Returns:
            Optional[Union[Session, Dict[str, Any]]]:
                - When deserialize=True: Session object
                - When deserialize=False: Session dictionary

        Raises:
            Exception: If an error occurs during retrieval.
        """
        sessions_table = self._get_table("sessions")
        record = RecordID(sessions_table, session_id)
        where = WhereClause()
        if user_id is not None:
            where = where.and_("user_id", user_id)
        if session_type == SessionType.AGENT:
            where = where.and_("agent", None, "!=")
        elif session_type == SessionType.TEAM:
            where = where.and_("team", None, "!=")
        elif session_type == SessionType.WORKFLOW:
            where = where.and_("workflow", None, "!=")
        where_clause, where_vars = where.build()
        query = dedent(f"""
            SELECT *
            FROM ONLY $record
            {where_clause}
        """)
        vars = {"record": record, **where_vars}
        raw = self._query_one(query, vars, dict)
        if raw is None or not deserialize:
            return raw

        return deserialize_session(session_type, raw)

    def get_sessions(
        self,
        session_type: Optional[SessionType] = None,
        user_id: Optional[str] = None,
        component_id: Optional[str] = None,
        session_name: Optional[str] = None,
        start_timestamp: Optional[int] = None,
        end_timestamp: Optional[int] = None,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
        deserialize: Optional[bool] = True,
    ) -> Union[List[Session], Tuple[List[Dict[str, Any]], int]]:
        r"""
        Get all sessions in the given table. Can filter by user_id and entity_id.

        Args:
            session_type (SessionType): The type of session to get.
            user_id (Optional[str]): The ID of the user to filter by.
            component_id (Optional[str]): The ID of the agent / team / workflow to filter by.
            session_name (Optional[str]): The name of the session to filter by.
            start_timestamp (Optional[int]): The start timestamp to filter by.
            end_timestamp (Optional[int]): The end timestamp to filter by.
            limit (Optional[int]): The maximum number of sessions to return. Defaults to None.
            page (Optional[int]): The page number to return. Defaults to None.
            sort_by (Optional[str]): The field to sort by. Defaults to None.
            sort_order (Optional[str]): The sort order. Defaults to None.
            deserialize (Optional[bool]): Whether to serialize the sessions. Defaults to True.

        Returns:
            Union[List[Session], Tuple[List[Dict], int]]:
                - When deserialize=True: List of Session objects
                - When deserialize=False: Tuple of (session dictionaries, total count)

        Raises:
            Exception: If an error occurs during retrieval.
        """
        table = self._get_table("sessions")
        # users_table = self._get_table("users", False) # Not used, commenting out for now.
        agents_table = self._get_table("agents", False)
        teams_table = self._get_table("teams", False)
        workflows_table = self._get_table("workflows", False)

        # -- Filters
        where = WhereClause()

        # user_id
        if user_id is not None:
            where = where.and_("user_id", user_id)

        # component_id
        if component_id is not None:
            if session_type == SessionType.AGENT:
                where = where.and_("agent", RecordID(agents_table, component_id))
            elif session_type == SessionType.TEAM:
                where = where.and_("team", RecordID(teams_table, component_id))
            elif session_type == SessionType.WORKFLOW:
                where = where.and_("workflow", RecordID(workflows_table, component_id))

        # session_name
        if session_name is not None:
            where = where.and_("session_name", session_name, "~")

        # start_timestamp
        if start_timestamp is not None:
            where = where.and_("start_timestamp", start_timestamp, ">=")

        # end_timestamp
        if end_timestamp is not None:
            where = where.and_("end_timestamp", end_timestamp, "<=")

        where_clause, where_vars = where.build()

        # Total count
        total_count = self._count(table, where_clause, where_vars)

        # Query
        order_limit_start_clause = order_limit_start(sort_by, sort_order, limit, page)
        query = dedent(f"""
            SELECT *
            FROM {table}
            {where_clause}
            {order_limit_start_clause}
        """)
        sessions_raw = self._query(query, where_vars, dict)
        converted_sessions_raw = [desurrealize_session(session, session_type) for session in sessions_raw]

        if not deserialize:
            return list(converted_sessions_raw), total_count

        if session_type is None:
            raise ValueError("session_type is required when deserialize=True")

        return deserialize_sessions(session_type, list(sessions_raw))

    def rename_session(
        self, session_id: str, session_type: SessionType, session_name: str, deserialize: Optional[bool] = True
    ) -> Optional[Union[Session, Dict[str, Any]]]:
        """
        Rename a session in the database.

        Args:
            session_id (str): The ID of the session to rename.
            session_type (SessionType): The type of session to rename.
            session_name (str): The new name for the session.
            deserialize (Optional[bool]): Whether to serialize the session. Defaults to True.

        Returns:
            Optional[Union[Session, Dict[str, Any]]]:
                - When deserialize=True: Session object
                - When deserialize=False: Session dictionary

        Raises:
            Exception: If an error occurs during renaming.
        """
        table = self._get_table("sessions")
        vars = {"record": RecordID(table, session_id), "name": session_name}

        # Query
        query = dedent("""
            UPDATE ONLY $record
            SET session_name = $name
        """)
        session_raw = self._query_one(query, vars, dict)

        if session_raw is None or not deserialize:
            return session_raw
        return deserialize_session(session_type, session_raw)

    def upsert_session(
        self, session: Session, deserialize: Optional[bool] = True
    ) -> Optional[Union[Session, Dict[str, Any]]]:
        """
        Insert or update a session in the database.

        Args:
            session (Session): The session data to upsert.
            deserialize (Optional[bool]): Whether to deserialize the session. Defaults to True.

        Returns:
            Optional[Union[Session, Dict[str, Any]]]:
                - When deserialize=True: Session object
                - When deserialize=False: Session dictionary

        Raises:
            Exception: If an error occurs during upsert.
        """
        session_type = get_session_type(session)
        table = self._get_table("sessions")
        session_raw = self._query_one(
            "UPSERT ONLY $record CONTENT $content",
            {
                "record": RecordID(table, session.session_id),
                "content": serialize_session(session, self.table_names),
            },
            dict,
        )
        if session_raw is None or not deserialize:
            return session_raw

        return deserialize_session(session_type, session_raw)

    def upsert_sessions(
        self, sessions: List[Session], deserialize: Optional[bool] = True
    ) -> List[Union[Session, Dict[str, Any]]]:
        """
        Bulk insert or update multiple sessions.

        Args:
            sessions (List[Session]): The list of session data to upsert.
            deserialize (Optional[bool]): Whether to deserialize the sessions. Defaults to True.

        Returns:
            List[Union[Session, Dict[str, Any]]]: List of upserted sessions

        Raises:
            Exception: If an error occurs during bulk upsert.
        """
        if not sessions:
            return []
        session_type = get_session_type(sessions[0])
        table = self._get_table("sessions")
        sessions_raw: List[Dict[str, Any]] = []
        for session in sessions:
            # UPSERT does only work for one record at a time
            session_raw = self._query_one(
                "UPSERT ONLY $record CONTENT $content",
                {
                    "record": RecordID(table, session.session_id),
                    "content": serialize_session(session, self.table_names),
                },
                dict,
            )
            if session_raw:
                sessions_raw.append(session_raw)
        if not deserialize:
            return list(sessions_raw)

        # wrapping with list because of:
        # Type "List[Session]" is not assignable to return type "List[Session | Dict[str, Any]]"
        # Consider switching from "list" to "Sequence" which is covariant
        return list(deserialize_sessions(session_type, sessions_raw))

    # --- Memory ---
    def clear_memories(self) -> None:
        """Delete all memories from the database.

        Raises:
            Exception: If an error occurs during deletion.
        """
        table = self._get_table("memories")
        _ = self.client.delete(table)

    def delete_user_memory(self, memory_id: str, user_id: Optional[str] = None) -> None:
        """Delete a user memory from the database.

        Args:
            memory_id (str): The ID of the memory to delete.
            user_id (Optional[str]): The ID of the user to filter by. Defaults to None.

        Returns:
            bool: True if deletion was successful, False otherwise.

        Raises:
            Exception: If an error occurs during deletion.
        """
        table = self._get_table("memories")
        mem_rec_id = RecordID(table, memory_id)
        if user_id is None:
            self.client.delete(mem_rec_id)
        else:
            user_rec_id = RecordID(self._get_table("users"), user_id)
            self.client.query(
                f"DELETE FROM {table} WHERE user = $user AND id = $memory",
                {"user": user_rec_id, "memory": mem_rec_id},
            )

    def delete_user_memories(self, memory_ids: List[str], user_id: Optional[str] = None) -> None:
        """Delete user memories from the database.

        Args:
            memory_ids (List[str]): The IDs of the memories to delete.
            user_id (Optional[str]): The ID of the user to filter by. Defaults to None.

        Raises:
            Exception: If an error occurs during deletion.
        """
        table = self._get_table("memories")
        records = [RecordID(table, memory_id) for memory_id in memory_ids]
        if user_id is None:
            _ = self.client.query(f"DELETE FROM {table} WHERE id IN $records", {"records": records})
        else:
            user_rec_id = RecordID(self._get_table("users"), user_id)
            _ = self.client.query(
                f"DELETE FROM {table} WHERE id IN $records AND user = $user", {"records": records, "user": user_rec_id}
            )

    def get_all_memory_topics(self, user_id: Optional[str] = None) -> List[str]:
        """Get all memory topics from the database.

        Args:
            user_id (Optional[str]): The ID of the user to filter by. Defaults to None.

        Returns:
            List[str]: List of memory topics.
        """
        table = self._get_table("memories")
        vars: dict[str, Any] = {}

        # Query
        if user_id is None:
            query = dedent(f"""
                RETURN (
                    SELECT
                        array::flatten(topics) as topics
                    FROM ONLY {table}
                    GROUP ALL
                ).topics.distinct();
            """)
        else:
            query = dedent(f"""
                RETURN (
                    SELECT
                        array::flatten(topics) as topics
                    FROM ONLY {table}
                    WHERE user = $user
                    GROUP ALL
                ).topics.distinct();
            """)
            vars["user"] = RecordID(self._get_table("users"), user_id)

        result = self._query(query, vars, str)
        return list(result)

    def get_user_memory(
        self, memory_id: str, deserialize: Optional[bool] = True, user_id: Optional[str] = None
    ) -> Optional[Union[UserMemory, Dict[str, Any]]]:
        """Get a memory from the database.

        Args:
            memory_id (str): The ID of the memory to get.
            deserialize (Optional[bool]): Whether to serialize the memory. Defaults to True.
            user_id (Optional[str]): The ID of the user to filter by. Defaults to None.

        Returns:
            Optional[Union[UserMemory, Dict[str, Any]]]:
                - When deserialize=True: UserMemory object
                - When deserialize=False: UserMemory dictionary

        Raises:
            Exception: If an error occurs during retrieval.
        """
        table_name = self._get_table("memories")
        record = RecordID(table_name, memory_id)
        vars = {"record": record}

        if user_id is None:
            query = "SELECT * FROM ONLY $record"
        else:
            query = "SELECT * FROM ONLY $record WHERE user = $user"
            vars["user"] = RecordID(self._get_table("users"), user_id)

        result = self._query_one(query, vars, dict)
        if result is None or not deserialize:
            return result
        return deserialize_user_memory(result)

    def get_user_memories(
        self,
        user_id: Optional[str] = None,
        agent_id: Optional[str] = None,
        team_id: Optional[str] = None,
        topics: Optional[List[str]] = None,
        search_content: Optional[str] = None,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
        deserialize: Optional[bool] = True,
    ) -> Union[List[UserMemory], Tuple[List[Dict[str, Any]], int]]:
        """Get all memories from the database as UserMemory objects.

        Args:
            user_id (Optional[str]): The ID of the user to filter by.
            agent_id (Optional[str]): The ID of the agent to filter by.
            team_id (Optional[str]): The ID of the team to filter by.
            topics (Optional[List[str]]): The topics to filter by.
            search_content (Optional[str]): The content to search for.
            limit (Optional[int]): The maximum number of memories to return.
            page (Optional[int]): The page number.
            sort_by (Optional[str]): The column to sort by.
            sort_order (Optional[str]): The order to sort by.
            deserialize (Optional[bool]): Whether to serialize the memories. Defaults to True.


        Returns:
            Union[List[UserMemory], Tuple[List[Dict[str, Any]], int]]:
                - When deserialize=True: List of UserMemory objects
                - When deserialize=False: Tuple of (memory dictionaries, total count)

        Raises:
            Exception: If an error occurs during retrieval.
        """
        table = self._get_table("memories")
        where = WhereClause()
        if user_id is not None:
            rec_id = RecordID(self._get_table("users"), user_id)
            where.and_("user", rec_id)
        if agent_id is not None:
            rec_id = RecordID(self._get_table("agents"), agent_id)
            where.and_("agent", rec_id)
        if team_id is not None:
            rec_id = RecordID(self._get_table("teams"), team_id)
            where.and_("team", rec_id)
        if topics is not None:
            where.and_("topics", topics, "CONTAINSANY")
        if search_content is not None:
            where.and_("memory", search_content, "~")
        where_clause, where_vars = where.build()

        # Total count
        total_count = self._count(table, where_clause, where_vars)

        # Query
        order_limit_start_clause = order_limit_start(sort_by, sort_order, limit, page)
        query = dedent(f"""
            SELECT *
            FROM {table}
            {where_clause}
            {order_limit_start_clause}
        """)
        result = self._query(query, where_vars, dict)
        if deserialize:
            return deserialize_user_memories(result)
        return [desurrealize_user_memory(x) for x in result], total_count

    def get_user_memory_stats(
        self,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        user_id: Optional[str] = None,
    ) -> Tuple[List[Dict[str, Any]], int]:
        """Get user memories stats.

        Args:
            limit (Optional[int]): The maximum number of user stats to return.
            page (Optional[int]): The page number.
            user_id (Optional[str]): The ID of the user to filter by. Defaults to None.

        Returns:
            Tuple[List[Dict[str, Any]], int]: A list of dictionaries containing user stats and total count.

        Example:
        (
            [
                {
                    "user_id": "123",
                    "total_memories": 10,
                    "last_memory_updated_at": 1714560000,
                },
            ],
            total_count: 1,
        )
        """
        memories_table_name = self._get_table("memories")
        where = WhereClause()

        if user_id is None:
            where.and_("!!user", True, "=")  # this checks that user is not falsy
        else:
            where.and_("user", RecordID(self._get_table("users"), user_id), "=")

        where_clause, where_vars = where.build()
        # Group
        group_clause = "GROUP BY user"
        # Order
        order_limit_start_clause = order_limit_start("last_memory_updated_at", "DESC", limit, page)
        # Total count
        total_count = (
            self._query_one(f"(SELECT user FROM {memories_table_name} GROUP BY user).map(|$x| $x.user).len()", {}, int)
            or 0
        )
        # Query
        query = dedent(f"""
            SELECT
                user,
                count(id) AS total_memories,
                time::max(updated_at) AS last_memory_updated_at
            FROM {memories_table_name}
            {where_clause}
            {group_clause}
            {order_limit_start_clause}
        """)
        result = self._query(query, where_vars, dict)

        # deserialize dates and RecordIDs
        for row in result:
            row["user_id"] = row["user"].id
            del row["user"]
            row["last_memory_updated_at"] = row["last_memory_updated_at"].timestamp()
            row["last_memory_updated_at"] = int(row["last_memory_updated_at"])

        return list(result), total_count

    def upsert_user_memory(
        self, memory: UserMemory, deserialize: Optional[bool] = True
    ) -> Optional[Union[UserMemory, Dict[str, Any]]]:
        """Upsert a user memory in the database.

        Args:
            memory (UserMemory): The user memory to upsert.
            deserialize (Optional[bool]): Whether to serialize the memory. Defaults to True.

        Returns:
            Optional[Union[UserMemory, Dict[str, Any]]]:
                - When deserialize=True: UserMemory object
                - When deserialize=False: UserMemory dictionary

        Raises:
            Exception: If an error occurs during upsert.
        """
        table = self._get_table("memories")
        user_table = self._get_table("users")
        if memory.memory_id:
            record = RecordID(table, memory.memory_id)
            query = "UPSERT ONLY $record CONTENT $content"
            result = self._query_one(
                query, {"record": record, "content": serialize_user_memory(memory, table, user_table)}, dict
            )
        else:
            query = f"CREATE ONLY {table} CONTENT $content"
            result = self._query_one(query, {"content": serialize_user_memory(memory, table, user_table)}, dict)
        if result is None:
            return None
        elif not deserialize:
            return desurrealize_user_memory(result)
        return deserialize_user_memory(result)

    def upsert_memories(
        self, memories: List[UserMemory], deserialize: Optional[bool] = True
    ) -> List[Union[UserMemory, Dict[str, Any]]]:
        """
        Bulk insert or update multiple memories in the database for improved performance.

        Args:
            memories (List[UserMemory]): The list of memories to upsert.
            deserialize (Optional[bool]): Whether to deserialize the memories. Defaults to True.

        Returns:
            List[Union[UserMemory, Dict[str, Any]]]: List of upserted memories

        Raises:
            Exception: If an error occurs during bulk upsert.
        """
        if not memories:
            return []
        table = self._get_table("memories")
        user_table_name = self._get_table("users")
        raw: list[dict] = []
        for memory in memories:
            if memory.memory_id:
                # UPSERT does only work for one record at a time
                session_raw = self._query_one(
                    "UPSERT ONLY $record CONTENT $content",
                    {
                        "record": RecordID(table, memory.memory_id),
                        "content": serialize_user_memory(memory, table, user_table_name),
                    },
                    dict,
                )
            else:
                session_raw = self._query_one(
                    f"CREATE ONLY {table} CONTENT $content",
                    {"content": serialize_user_memory(memory, table, user_table_name)},
                    dict,
                )
            if session_raw is not None:
                raw.append(session_raw)
        if raw is None or not deserialize:
            return [desurrealize_user_memory(x) for x in raw]
        # wrapping with list because of:
        # Type "List[Session]" is not assignable to return type "List[Session | Dict[str, Any]]"
        # Consider switching from "list" to "Sequence" which is covariant
        return list(deserialize_user_memories(raw))

    # --- Metrics ---
    def get_metrics(
        self,
        starting_date: Optional[date] = None,
        ending_date: Optional[date] = None,
    ) -> Tuple[List[Dict[str, Any]], Optional[int]]:
        """Get all metrics matching the given date range.

        Args:
            starting_date (Optional[date]): The starting date to filter metrics by.
            ending_date (Optional[date]): The ending date to filter metrics by.

        Returns:
            Tuple[List[dict], Optional[int]]: A tuple containing the metrics and the timestamp of the latest update.

        Raises:
            Exception: If an error occurs during retrieval.
        """
        table = self._get_table("metrics")

        where = WhereClause()

        # starting_date - need to convert date to datetime for comparison
        if starting_date is not None:
            starting_datetime = datetime.combine(starting_date, datetime.min.time()).replace(tzinfo=timezone.utc)
            where = where.and_("date", starting_datetime, ">=")

        # ending_date - need to convert date to datetime for comparison
        if ending_date is not None:
            ending_datetime = datetime.combine(ending_date, datetime.min.time()).replace(tzinfo=timezone.utc)
            where = where.and_("date", ending_datetime, "<=")

        where_clause, where_vars = where.build()

        # Query
        query = dedent(f"""
            SELECT *
            FROM {table}
            {where_clause}
            ORDER BY date ASC
        """)

        results = self._query(query, where_vars, dict)

        # Get the latest updated_at from all results
        latest_update = None
        if results:
            # Find the maximum updated_at timestamp
            latest_update = max(int(r["updated_at"].timestamp()) for r in results)

            # Transform results to match expected format
            transformed_results = []
            for r in results:
                transformed = dict(r)

                # Convert RecordID to string
                if hasattr(transformed.get("id"), "id"):
                    transformed["id"] = transformed["id"].id
                elif isinstance(transformed.get("id"), RecordID):
                    transformed["id"] = str(transformed["id"].id)

                # Convert datetime objects to Unix timestamps
                if isinstance(transformed.get("created_at"), datetime):
                    transformed["created_at"] = int(transformed["created_at"].timestamp())
                if isinstance(transformed.get("updated_at"), datetime):
                    transformed["updated_at"] = int(transformed["updated_at"].timestamp())
                if isinstance(transformed.get("date"), datetime):
                    transformed["date"] = int(transformed["date"].timestamp())

                transformed_results.append(transformed)

            return transformed_results, latest_update

        return [], latest_update

    def calculate_metrics(self) -> Optional[List[Dict[str, Any]]]:  # More specific return type
        """Calculate metrics for all dates without complete metrics.

        Returns:
            Optional[List[Dict[str, Any]]]: The calculated metrics.

        Raises:
            Exception: If an error occurs during metrics calculation.
        """
        try:
            table = self._get_table("metrics")  # Removed create_table_if_not_found parameter

            starting_date = get_metrics_calculation_starting_date(self.client, table, self.get_sessions)

            if starting_date is None:
                log_info("No session data found. Won't calculate metrics.")
                return None

            dates_to_process = get_dates_to_calculate_metrics_for(starting_date)
            if not dates_to_process:
                log_info("Metrics already calculated for all relevant dates.")
                return None

            start_timestamp = datetime.combine(dates_to_process[0], datetime.min.time()).replace(tzinfo=timezone.utc)
            end_timestamp = datetime.combine(dates_to_process[-1] + timedelta(days=1), datetime.min.time()).replace(
                tzinfo=timezone.utc
            )

            sessions = get_all_sessions_for_metrics_calculation(
                self.client, self._get_table("sessions"), start_timestamp, end_timestamp
            )

            all_sessions_data = fetch_all_sessions_data(
                sessions=sessions,  # Added parameter name for clarity
                dates_to_process=dates_to_process,
                start_timestamp=int(start_timestamp.timestamp()),  # This expects int
            )
            if not all_sessions_data:
                log_info("No new session data found. Won't calculate metrics.")
                return None

            metrics_records = []

            for date_to_process in dates_to_process:
                date_key = date_to_process.isoformat()
                sessions_for_date = all_sessions_data.get(date_key, {})

                # Skip dates with no sessions
                if not any(len(sessions) > 0 for sessions in sessions_for_date.values()):
                    continue

                metrics_record = calculate_date_metrics(date_to_process, sessions_for_date)
                metrics_records.append(metrics_record)

            results = []  # Initialize before the if block
            if metrics_records:
                results = bulk_upsert_metrics(self.client, table, metrics_records)

            log_debug("Updated metrics calculations")
            return results

        except Exception as e:
            log_error(f"Exception refreshing metrics: {e}")
            raise e

    # --- Knowledge ---
    def clear_knowledge(self) -> None:
        """Delete all knowledge rows from the database.

        Raises:
            Exception: If an error occurs during deletion.
        """
        table = self._get_table("knowledge")
        _ = self.client.delete(table)

    def delete_knowledge_content(self, id: str):
        """Delete a knowledge row from the database.

        Args:
            id (str): The ID of the knowledge row to delete.
        """
        table = self._get_table("knowledge")
        self.client.delete(RecordID(table, id))

    def get_knowledge_content(self, id: str) -> Optional[KnowledgeRow]:
        """Get a knowledge row from the database.

        Args:
            id (str): The ID of the knowledge row to get.

        Returns:
            Optional[KnowledgeRow]: The knowledge row, or None if it doesn't exist.
        """
        table = self._get_table("knowledge")
        record_id = RecordID(table, id)
        raw = self._query_one("SELECT * FROM ONLY $record_id", {"record_id": record_id}, dict)
        return deserialize_knowledge_row(raw) if raw else None

    def get_knowledge_contents(
        self,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
    ) -> Tuple[List[KnowledgeRow], int]:
        """Get all knowledge contents from the database.

        Args:
            limit (Optional[int]): The maximum number of knowledge contents to return.
            page (Optional[int]): The page number.
            sort_by (Optional[str]): The column to sort by.
            sort_order (Optional[str]): The order to sort by.

        Returns:
            Tuple[List[KnowledgeRow], int]: The knowledge contents and total count.

        Raises:
            Exception: If an error occurs during retrieval.
        """
        table = self._get_table("knowledge")
        where = WhereClause()
        where_clause, where_vars = where.build()

        # Total count
        total_count = self._count(table, where_clause, where_vars)

        # Query
        order_limit_start_clause = order_limit_start(sort_by, sort_order, limit, page)
        query = dedent(f"""
            SELECT *
            FROM {table}
            {where_clause}
            {order_limit_start_clause}
        """)
        result = self._query(query, where_vars, dict)
        return [deserialize_knowledge_row(row) for row in result], total_count

    def upsert_knowledge_content(self, knowledge_row: KnowledgeRow) -> Optional[KnowledgeRow]:
        """Upsert knowledge content in the database.

        Args:
            knowledge_row (KnowledgeRow): The knowledge row to upsert.

        Returns:
            Optional[KnowledgeRow]: The upserted knowledge row, or None if the operation fails.
        """
        knowledge_table_name = self._get_table("knowledge")
        record = RecordID(knowledge_table_name, knowledge_row.id)
        query = "UPSERT ONLY $record CONTENT $content"
        result = self._query_one(
            query, {"record": record, "content": serialize_knowledge_row(knowledge_row, knowledge_table_name)}, dict
        )
        return deserialize_knowledge_row(result) if result else None

    # --- Evals ---
    def clear_evals(self) -> None:
        """Delete all eval rows from the database.

        Raises:
            Exception: If an error occurs during deletion.
        """
        table = self._get_table("evals")
        _ = self.client.delete(table)

    def create_eval_run(self, eval_run: EvalRunRecord) -> Optional[EvalRunRecord]:
        """Create an EvalRunRecord in the database.

        Args:
            eval_run (EvalRunRecord): The eval run to create.

        Returns:
            Optional[EvalRunRecord]: The created eval run, or None if the operation fails.

        Raises:
            Exception: If an error occurs during creation.
        """
        table = self._get_table("evals")
        rec_id = RecordID(table, eval_run.run_id)
        query = "CREATE ONLY $record CONTENT $content"
        result = self._query_one(
            query, {"record": rec_id, "content": serialize_eval_run_record(eval_run, self.table_names)}, dict
        )
        return deserialize_eval_run_record(result) if result else None

    def delete_eval_runs(self, eval_run_ids: List[str]) -> None:
        """Delete multiple eval runs from the database.

        Args:
            eval_run_ids (List[str]): List of eval run IDs to delete.
        """
        table = self._get_table("evals")
        records = [RecordID(table, id) for id in eval_run_ids]
        _ = self.client.query(f"DELETE FROM {table} WHERE id IN $records", {"records": records})

    def get_eval_run(
        self, eval_run_id: str, deserialize: Optional[bool] = True
    ) -> Optional[Union[EvalRunRecord, Dict[str, Any]]]:
        """Get an eval run from the database.

        Args:
            eval_run_id (str): The ID of the eval run to get.
            deserialize (Optional[bool]): Whether to serialize the eval run. Defaults to True.

        Returns:
            Optional[Union[EvalRunRecord, Dict[str, Any]]]:
                - When deserialize=True: EvalRunRecord object
                - When deserialize=False: EvalRun dictionary

        Raises:
            Exception: If an error occurs during retrieval.
        """
        table = self._get_table("evals")
        record = RecordID(table, eval_run_id)
        result = self._query_one("SELECT * FROM ONLY $record", {"record": record}, dict)
        if not result or not deserialize:
            return desurrealize_eval_run_record(result) if result is not None else None
        return deserialize_eval_run_record(result)

    def get_eval_runs(
        self,
        limit: Optional[int] = None,
        page: Optional[int] = None,
        sort_by: Optional[str] = None,
        sort_order: Optional[str] = None,
        agent_id: Optional[str] = None,
        team_id: Optional[str] = None,
        workflow_id: Optional[str] = None,
        model_id: Optional[str] = None,
        filter_type: Optional[EvalFilterType] = None,
        eval_type: Optional[List[EvalType]] = None,
        deserialize: Optional[bool] = True,
    ) -> Union[List[EvalRunRecord], Tuple[List[Dict[str, Any]], int]]:
        """Get all eval runs from the database.

        Args:
            limit (Optional[int]): The maximum number of eval runs to return.
            page (Optional[int]): The page number to return.
            sort_by (Optional[str]): The field to sort by.
            sort_order (Optional[str]): The order to sort by.
            agent_id (Optional[str]): The ID of the agent to filter by.
            team_id (Optional[str]): The ID of the team to filter by.
            workflow_id (Optional[str]): The ID of the workflow to filter by.
            model_id (Optional[str]): The ID of the model to filter by.
            eval_type (Optional[List[EvalType]]): The type of eval to filter by.
            filter_type (Optional[EvalFilterType]): The type of filter to apply.
            deserialize (Optional[bool]): Whether to serialize the eval runs. Defaults to True.

        Returns:
            Union[List[EvalRunRecord], Tuple[List[Dict[str, Any]], int]]:
                - When deserialize=True: List of EvalRunRecord objects
                - When deserialize=False: List of eval run dictionaries and the total count

        Raises:
            Exception: If there is an error getting the eval runs.
        """
        table = self._get_table("evals")

        where = WhereClause()
        if filter_type is not None:
            if filter_type == EvalFilterType.AGENT:
                where.and_("agent", RecordID(self._get_table("agents"), agent_id))
            elif filter_type == EvalFilterType.TEAM:
                where.and_("team", RecordID(self._get_table("teams"), team_id))
            elif filter_type == EvalFilterType.WORKFLOW:
                where.and_("workflow", RecordID(self._get_table("workflows"), workflow_id))
        if model_id is not None:
            where.and_("model_id", model_id)
        if eval_type is not None:
            where.and_("eval_type", eval_type)
        where_clause, where_vars = where.build()

        # Order
        order_limit_start_clause = order_limit_start(sort_by, sort_order, limit, page)

        # Total count
        total_count = self._count(table, where_clause, where_vars)

        # Query
        query = dedent(f"""
            SELECT *
            FROM {table}
            {where_clause}
            {order_limit_start_clause}
        """)
        result = self._query(query, where_vars, dict)

        if not deserialize:
            return list(result), total_count
        return [deserialize_eval_run_record(x) for x in result]

    def rename_eval_run(
        self, eval_run_id: str, name: str, deserialize: Optional[bool] = True
    ) -> Optional[Union[EvalRunRecord, Dict[str, Any]]]:
        """Update the name of an eval run in the database.

        Args:
            eval_run_id (str): The ID of the eval run to update.
            name (str): The new name of the eval run.
            deserialize (Optional[bool]): Whether to serialize the eval run. Defaults to True.

        Returns:
            Optional[Union[EvalRunRecord, Dict[str, Any]]]:
                - When deserialize=True: EvalRunRecord object
                - When deserialize=False: EvalRun dictionary

        Raises:
            Exception: If there is an error updating the eval run.
        """
        table = self._get_table("evals")
        vars = {"record": RecordID(table, eval_run_id), "name": name}

        # Query
        query = dedent("""
            UPDATE ONLY $record
            SET name = $name
        """)
        raw = self._query_one(query, vars, dict)

        if not raw or not deserialize:
            return raw
        return deserialize_eval_run_record(raw)



from typing import List, Optional

from agno.knowledge.chunking.fixed import FixedSizeChunking
from agno.knowledge.chunking.strategy import ChunkingStrategy, ChunkingStrategyType
from agno.knowledge.document import Document
from agno.knowledge.reader.base import Reader
from agno.knowledge.types import ContentType
from agno.utils.log import log_debug, log_info

try:
    import wikipedia  # noqa: F401
except ImportError:
    raise ImportError("The `wikipedia` package is not installed. Please install it via `pip install wikipedia`.")


class WikipediaReader(Reader):
    auto_suggest: bool = True

    def __init__(
        self, chunking_strategy: Optional[ChunkingStrategy] = FixedSizeChunking(), auto_suggest: bool = True, **kwargs
    ):
        super().__init__(chunking_strategy=chunking_strategy, **kwargs)
        self.auto_suggest = auto_suggest

    @classmethod
    def get_supported_chunking_strategies(self) -> List[ChunkingStrategyType]:
        """Get the list of supported chunking strategies for Wikipedia readers."""
        return [
            ChunkingStrategyType.FIXED_SIZE_CHUNKER,
            ChunkingStrategyType.AGENTIC_CHUNKER,
            ChunkingStrategyType.DOCUMENT_CHUNKER,
            ChunkingStrategyType.RECURSIVE_CHUNKER,
            ChunkingStrategyType.SEMANTIC_CHUNKER,
        ]

    @classmethod
    def get_supported_content_types(self) -> List[ContentType]:
        return [ContentType.TOPIC]

    def read(self, topic: str) -> List[Document]:
        log_debug(f"Reading Wikipedia topic: {topic}")
        summary = None
        try:
            summary = wikipedia.summary(topic, auto_suggest=self.auto_suggest)

        except wikipedia.exceptions.PageError:
            summary = None
            log_info("PageError: Page not found.")

        # Only create Document if we successfully got a summary
        if summary:
            return [
                Document(
                    name=topic,
                    meta_data={"topic": topic},
                    content=summary,
                )
            ]
        return []



"""Table schemas and related utils used by the SingleStoreDb class"""

from typing import Any

try:
    from sqlalchemy.types import JSON, BigInteger, Boolean, Date, String, Text
except ImportError:
    raise ImportError("`sqlalchemy` not installed. Please install it using `pip install sqlalchemy`")

SESSION_TABLE_SCHEMA = {
    "session_id": {"type": lambda: String(128), "nullable": False},
    "session_type": {"type": lambda: String(20), "nullable": False, "index": True},
    "agent_id": {"type": lambda: String(128), "nullable": True},
    "team_id": {"type": lambda: String(128), "nullable": True},
    "workflow_id": {"type": lambda: String(128), "nullable": True},
    "user_id": {"type": lambda: String(128), "nullable": True},
    "session_data": {"type": JSON, "nullable": True},
    "agent_data": {"type": JSON, "nullable": True},
    "team_data": {"type": JSON, "nullable": True},
    "workflow_data": {"type": JSON, "nullable": True},
    "metadata": {"type": JSON, "nullable": True},
    "runs": {"type": JSON, "nullable": True},
    "summary": {"type": JSON, "nullable": True},
    "created_at": {"type": BigInteger, "nullable": False, "index": True},
    "updated_at": {"type": BigInteger, "nullable": True},
    "_unique_constraints": [
        {
            "name": "uq_session_id",
            "columns": ["session_id"],
        },
    ],
}

USER_MEMORY_TABLE_SCHEMA = {
    "memory_id": {"type": lambda: String(128), "primary_key": True, "nullable": False},
    "memory": {"type": JSON, "nullable": False},
    "input": {"type": Text, "nullable": True},
    "agent_id": {"type": lambda: String(128), "nullable": True},
    "team_id": {"type": lambda: String(128), "nullable": True},
    "user_id": {"type": lambda: String(128), "nullable": True, "index": True},
    "topics": {"type": JSON, "nullable": True},
    "updated_at": {"type": BigInteger, "nullable": True, "index": True},
}

EVAL_TABLE_SCHEMA = {
    "run_id": {"type": lambda: String(128), "primary_key": True, "nullable": False},
    "eval_type": {"type": lambda: String(50), "nullable": False},
    "eval_data": {"type": JSON, "nullable": False},
    "eval_input": {"type": JSON, "nullable": False},
    "name": {"type": lambda: String(255), "nullable": True},
    "agent_id": {"type": lambda: String(128), "nullable": True},
    "team_id": {"type": lambda: String(128), "nullable": True},
    "workflow_id": {"type": lambda: String(128), "nullable": True},
    "model_id": {"type": lambda: String(128), "nullable": True},
    "model_provider": {"type": lambda: String(50), "nullable": True},
    "evaluated_component_name": {"type": lambda: String(255), "nullable": True},
    "created_at": {"type": BigInteger, "nullable": False, "index": True},
    "updated_at": {"type": BigInteger, "nullable": True},
}

KNOWLEDGE_TABLE_SCHEMA = {
    "id": {"type": lambda: String(128), "primary_key": True, "nullable": False},
    "name": {"type": lambda: String(255), "nullable": False},
    "description": {"type": Text, "nullable": False},
    "metadata": {"type": JSON, "nullable": True},
    "type": {"type": lambda: String(50), "nullable": True},
    "size": {"type": BigInteger, "nullable": True},
    "linked_to": {"type": lambda: String(128), "nullable": True},
    "access_count": {"type": BigInteger, "nullable": True},
    "status": {"type": lambda: String(50), "nullable": True},
    "status_message": {"type": Text, "nullable": True},
    "created_at": {"type": BigInteger, "nullable": True},
    "updated_at": {"type": BigInteger, "nullable": True},
    "external_id": {"type": lambda: String(128), "nullable": True},
}

METRICS_TABLE_SCHEMA = {
    "id": {"type": lambda: String(128), "primary_key": True, "nullable": False},
    "agent_runs_count": {"type": BigInteger, "nullable": False, "default": 0},
    "team_runs_count": {"type": BigInteger, "nullable": False, "default": 0},
    "workflow_runs_count": {"type": BigInteger, "nullable": False, "default": 0},
    "agent_sessions_count": {"type": BigInteger, "nullable": False, "default": 0},
    "team_sessions_count": {"type": BigInteger, "nullable": False, "default": 0},
    "workflow_sessions_count": {"type": BigInteger, "nullable": False, "default": 0},
    "users_count": {"type": BigInteger, "nullable": False, "default": 0},
    "token_metrics": {"type": JSON, "nullable": False, "default": {}},
    "model_metrics": {"type": JSON, "nullable": False, "default": {}},
    "date": {"type": Date, "nullable": False, "index": True},
    "aggregation_period": {"type": lambda: String(20), "nullable": False, "index": True},
    "created_at": {"type": BigInteger, "nullable": False},
    "updated_at": {"type": BigInteger, "nullable": True},
    "completed": {"type": Boolean, "nullable": False, "default": False},
}


def get_table_schema_definition(table_type: str) -> dict[str, Any]:
    """
    Get the expected schema definition for the given table.
    Args:
        table_type (str): The type of table to get the schema for.
    Returns:
        Dict[str, Any]: Dictionary containing column definitions for the table
    """
    schemas = {
        "sessions": SESSION_TABLE_SCHEMA,
        "evals": EVAL_TABLE_SCHEMA,
        "metrics": METRICS_TABLE_SCHEMA,
        "memories": USER_MEMORY_TABLE_SCHEMA,
        "knowledge": KNOWLEDGE_TABLE_SCHEMA,
    }
    schema = schemas.get(table_type, {})

    if not schema:
        raise ValueError(f"Unknown table type: {table_type}")

    return schema  # type: ignore[return-value]



import json
from typing import Any, Dict, List, Optional

from agno.tools.toolkit import Toolkit
from agno.utils.log import log_debug, log_error, log_info, log_warning

try:
    from memori import Memori, create_memory_tool
except ImportError:
    raise ImportError("`memorisdk` package not found. Please install it with `pip install memorisdk`")


class MemoriTools(Toolkit):
    """
    Memori ToolKit for Agno Agents and Teams, providing persistent memory capabilities.

    This toolkit integrates Memori's memory system with Agno, allowing Agents and Teams to:
    - Store and retrieve conversation history
    - Search through past interactions
    - Maintain user preferences and context
    - Build long-term memory across sessions

    Requirements:
        - pip install memorisdk
        - Database connection string (SQLite, PostgreSQL, etc.)

    Example:
        ```python
        from agno.tools.memori import MemoriTools

        # Initialize with SQLite (default)
        memori_tools = MemoriTools(
            database_connect="sqlite:///agent_memory.db",
            namespace="my_agent",
            auto_ingest=True  # Automatically ingest conversations
        )

        # Add to agent
        agent = Agent(
            model=OpenAIChat(),
            tools=[memori_tools],
            description="An AI assistant with persistent memory"
        )
        ```
    """

    def __init__(
        self,
        database_connect: Optional[str] = None,
        namespace: Optional[str] = None,
        conscious_ingest: bool = True,
        auto_ingest: bool = True,
        verbose: bool = False,
        config: Optional[Dict[str, Any]] = None,
        auto_enable: bool = True,
        enable_search_memory: bool = True,
        enable_record_conversation: bool = True,
        enable_get_memory_stats: bool = True,
        all: bool = False,
        **kwargs,
    ):
        """
        Initialize Memori toolkit.

        Args:
            database_connect: Database connection string (e.g., "sqlite:///memory.db")
            namespace: Namespace for organizing memories (e.g., "agent_v1", "user_session")
            conscious_ingest: Whether to use conscious memory ingestion
            auto_ingest: Whether to automatically ingest conversations into memory
            verbose: Enable verbose logging from Memori
            config: Additional Memori configuration
            auto_enable: Automatically enable the memory system on initialization
            **kwargs: Additional arguments passed to Toolkit base class
        """
        tools: List[Any] = []
        if all or enable_search_memory:
            tools.append(self.search_memory)
        if all or enable_record_conversation:
            tools.append(self.record_conversation)
        if all or enable_get_memory_stats:
            tools.append(self.get_memory_stats)

        super().__init__(name="memori_tools", tools=tools, **kwargs)

        # Set default database connection if not provided
        if not database_connect:
            sqlite_db = "sqlite:///agno_memori_memory.db"
            log_info(f"No database connection provided, using default SQLite database at {sqlite_db}")
            database_connect = sqlite_db

        self.database_connect = database_connect
        self.namespace = namespace or "agno_default"
        self.conscious_ingest = conscious_ingest
        self.auto_ingest = auto_ingest
        self.verbose = verbose
        self.config = config or {}

        try:
            # Initialize Memori memory system
            log_debug(f"Initializing Memori with database: {self.database_connect}")
            self.memory_system = Memori(
                database_connect=self.database_connect,
                conscious_ingest=self.conscious_ingest,
                auto_ingest=self.auto_ingest,
                verbose=self.verbose,
                namespace=self.namespace,
                **self.config,
            )

            # Enable the memory system if auto_enable is True
            if auto_enable:
                self.memory_system.enable()
                log_debug("Memori memory system enabled")

            # Create the memory tool for internal use
            self._memory_tool = create_memory_tool(self.memory_system)

        except Exception as e:
            log_error(f"Failed to initialize Memori: {e}")
            raise ConnectionError("Failed to initialize Memori memory system") from e

    def search_memory(
        self,
        query: str,
        limit: Optional[int] = None,
    ) -> str:
        """
        Search the Agent's memory for past conversations and information.

        This performs semantic search across all stored memories to find
        relevant information based on the provided query.

        Args:
            query: What to search for in memory (e.g., "past conversations about AI", "user preferences")
            limit: Maximum number of results to return (optional)

        Returns:
            str: JSON-encoded search results or error message

        Example:
            search_memory("user's favorite programming languages")
            search_memory("previous discussions about machine learning")
        """
        try:
            if not query.strip():
                return json.dumps({"error": "Please provide a search query"})

            log_debug(f"Searching memory for: {query}")

            # Execute search using Memori's memory tool
            result = self._memory_tool.execute(query=query.strip())

            if result:
                # If limit is specified, truncate results
                if limit and isinstance(result, list):
                    result = result[:limit]

                return json.dumps(
                    {
                        "success": True,
                        "query": query,
                        "results": result,
                        "count": len(result) if isinstance(result, list) else 1,
                    }
                )
            else:
                return json.dumps(
                    {
                        "success": True,
                        "query": query,
                        "results": [],
                        "count": 0,
                        "message": "No relevant memories found",
                    }
                )

        except Exception as e:
            log_error(f"Error searching memory: {e}")
            return json.dumps({"success": False, "error": f"Memory search error: {str(e)}"})

    def record_conversation(self, content: str) -> str:
        """
        Add important information or facts to memory.

        Use this tool to store important information, user preferences, facts, or context that should be remembered
        for future conversations.

        Args:
            content: The information/facts to store in memory

        Returns:
            str: Success message or error details

        Example:
            record_conversation("User prefers Python over JavaScript")
            record_conversation("User is working on an e-commerce project using Django")
            record_conversation("User's name is John and they live in NYC")
        """
        try:
            if not content.strip():
                return json.dumps({"success": False, "error": "Content cannot be empty"})

            log_debug(f"Adding conversation: {content}")

            # Extract the actual AI response from the agent's conversation history
            ai_output = "I've noted this information and will remember it."

            self.memory_system.record_conversation(user_input=content, ai_output=str(ai_output))
            return json.dumps(
                {
                    "success": True,
                    "message": "Memory added successfully via conversation recording",
                    "content_length": len(content),
                }
            )

        except Exception as e:
            log_error(f"Error adding memory: {e}")
            return json.dumps({"success": False, "error": f"Failed to add memory: {str(e)}"})

    def get_memory_stats(
        self,
    ) -> str:
        """
        Get statistics about the memory system.

        Returns information about the current state of the memory system,
        including total memories, memory distribution by retention type
        (short-term vs long-term), and system configuration.

        Returns:
            str: JSON-encoded memory statistics

        Example:
            Returns statistics like:
            {
                "success": true,
                "total_memories": 42,
                "memories_by_retention": {
                    "short_term": 5,
                    "long_term": 37
                },
                "namespace": "my_agent",
                "conscious_ingest": true,
                "auto_ingest": true,
                "memory_system_enabled": true
            }
        """
        try:
            log_debug("Retrieving memory statistics")

            # Base stats about the system configuration
            stats = {
                "success": True,
                "namespace": self.namespace,
                "database_connect": self.database_connect,
                "conscious_ingest": self.conscious_ingest,
                "auto_ingest": self.auto_ingest,
                "verbose": self.verbose,
                "memory_system_enabled": hasattr(self.memory_system, "_enabled") and self.memory_system._enabled,
            }

            # Get Memori's built-in memory statistics
            try:
                if hasattr(self.memory_system, "get_memory_stats"):
                    # Use the get_memory_stats method as shown in the example
                    memori_stats = self.memory_system.get_memory_stats()

                    # Add the Memori-specific stats to our response
                    if isinstance(memori_stats, dict):
                        # Include total memories
                        if "total_memories" in memori_stats:
                            stats["total_memories"] = memori_stats["total_memories"]

                        # Include memory distribution by retention type
                        if "memories_by_retention" in memori_stats:
                            stats["memories_by_retention"] = memori_stats["memories_by_retention"]

                            # Also add individual counts for convenience
                            retention_info = memori_stats["memories_by_retention"]
                            stats["short_term_memories"] = retention_info.get("short_term", 0)
                            stats["long_term_memories"] = retention_info.get("long_term", 0)

                        # Include any other available stats
                        for key, value in memori_stats.items():
                            if key not in stats:
                                stats[key] = value

                    log_debug(
                        f"Retrieved memory stats: total={stats.get('total_memories', 0)}, "
                        f"short_term={stats.get('short_term_memories', 0)}, "
                        f"long_term={stats.get('long_term_memories', 0)}"
                    )

                else:
                    log_debug("get_memory_stats method not available, providing basic stats only")
                    stats["total_memories"] = 0
                    stats["memories_by_retention"] = {"short_term": 0, "long_term": 0}
                    stats["short_term_memories"] = 0
                    stats["long_term_memories"] = 0

            except Exception as e:
                log_debug(f"Could not retrieve detailed memory stats: {e}")
                # Provide basic stats if detailed stats fail
                stats["total_memories"] = 0
                stats["memories_by_retention"] = {"short_term": 0, "long_term": 0}
                stats["short_term_memories"] = 0
                stats["long_term_memories"] = 0
                stats["stats_warning"] = "Detailed memory statistics not available"

            return json.dumps(stats)

        except Exception as e:
            log_error(f"Error getting memory stats: {e}")
            return json.dumps({"success": False, "error": f"Failed to get memory statistics: {str(e)}"})

    def enable_memory_system(self) -> bool:
        """Enable the Memori memory system."""
        try:
            self.memory_system.enable()
            log_debug("Memori memory system enabled")
            return True
        except Exception as e:
            log_error(f"Failed to enable memory system: {e}")
            return False

    def disable_memory_system(self) -> bool:
        """Disable the Memori memory system."""
        try:
            if hasattr(self.memory_system, "disable"):
                self.memory_system.disable()
                log_debug("Memori memory system disabled")
                return True
            else:
                log_warning("Memory system disable method not available")
                return False
        except Exception as e:
            log_error(f"Failed to disable memory system: {e}")
            return False



import json
from time import time
from typing import Any, Dict, List, Optional, Sequence, Union

from uuid import uuid4

from pydantic import BaseModel, ConfigDict, Field

from agno.media import Audio, File, Image, Video
from agno.models.metrics import Metrics
from agno.utils.log import log_debug, log_error, log_info, log_warning


class MessageReferences(BaseModel):
    """References added to user message"""

    # The query used to retrieve the references.
    query: str
    # References (from the vector database or function calls)
    references: Optional[List[Union[Dict[str, Any], str]]] = None
    # Time taken to retrieve the references.
    time: Optional[float] = None


class UrlCitation(BaseModel):
    """URL of the citation"""

    url: Optional[str] = None
    title: Optional[str] = None


class DocumentCitation(BaseModel):
    """Document of the citation"""

    document_title: Optional[str] = None
    cited_text: Optional[str] = None
    file_name: Optional[str] = None


class Citations(BaseModel):
    """Citations for the message"""

    # Raw citations from the model
    raw: Optional[Any] = None

    # URLs of the citations.
    urls: Optional[List[UrlCitation]] = None

    # Document Citations
    documents: Optional[List[DocumentCitation]] = None


class Message(BaseModel):
    """Message sent to the Model"""

    id: str = Field(default_factory=lambda: str(uuid4()))

    # The role of the message author.
    # One of system, user, assistant, or tool.
    role: str
    # The contents of the message.
    content: Optional[Union[List[Any], str]] = None
    # An optional name for the participant.
    # Provides the model information to differentiate between participants of the same role.
    name: Optional[str] = None
    # Tool call that this message is responding to.
    tool_call_id: Optional[str] = None
    # The tool calls generated by the model, such as function calls.
    tool_calls: Optional[List[Dict[str, Any]]] = None

    # Additional modalities
    audio: Optional[Sequence[Audio]] = None
    images: Optional[Sequence[Image]] = None
    videos: Optional[Sequence[Video]] = None
    files: Optional[Sequence[File]] = None

    # Output from the models
    audio_output: Optional[Audio] = None
    image_output: Optional[Image] = None
    video_output: Optional[Video] = None
    file_output: Optional[File] = None

    # The thinking content from the model
    redacted_reasoning_content: Optional[str] = None

    # Data from the provider we might need on subsequent messages
    provider_data: Optional[Dict[str, Any]] = None

    # Citations received from the model
    citations: Optional[Citations] = None

    # --- Data not sent to the Model API ---
    # The reasoning content from the model
    reasoning_content: Optional[str] = None
    # The name of the tool called
    tool_name: Optional[str] = None
    # Arguments passed to the tool
    tool_args: Optional[Any] = None
    # The error of the tool call
    tool_call_error: Optional[bool] = None
    # If True, the agent will stop executing after this tool call.
    stop_after_tool_call: bool = False
    # When True, the message will be added to the agent's memory.
    add_to_agent_memory: bool = True
    # This flag is enabled when a message is fetched from the agent's memory.
    from_history: bool = False
    # Metrics for the message.
    metrics: Metrics = Field(default_factory=Metrics)
    # The references added to the message for RAG
    references: Optional[MessageReferences] = None
    # The Unix timestamp the message was created.
    created_at: int = Field(default_factory=lambda: int(time()))

    model_config = ConfigDict(extra="allow", populate_by_name=True, arbitrary_types_allowed=True)

    def get_content_string(self) -> str:
        """Returns the content as a string."""
        if isinstance(self.content, str):
            return self.content
        if isinstance(self.content, list):
            if len(self.content) > 0 and isinstance(self.content[0], dict) and "text" in self.content[0]:
                return self.content[0].get("text", "")
            else:
                return json.dumps(self.content)
        return ""

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "Message":
        # Handle image reconstruction properly
        if "images" in data and data["images"]:
            reconstructed_images = []
            for i, img_data in enumerate(data["images"]):
                if isinstance(img_data, dict):
                    # If content is base64, decode it back to bytes
                    if "content" in img_data and isinstance(img_data["content"], str):
                        reconstructed_images.append(
                            Image.from_base64(
                                img_data["content"],
                                id=img_data.get("id"),
                                mime_type=img_data.get("mime_type"),
                                format=img_data.get("format"),
                            )
                        )
                    else:
                        # Regular image (filepath/url)
                        reconstructed_images.append(Image(**img_data))
                else:
                    reconstructed_images.append(img_data)
            data["images"] = reconstructed_images

        # Handle audio reconstruction properly
        if "audio" in data and data["audio"]:
            reconstructed_audio = []
            for i, aud_data in enumerate(data["audio"]):
                if isinstance(aud_data, dict):
                    # If content is base64, decode it back to bytes
                    if "content" in aud_data and isinstance(aud_data["content"], str):
                        reconstructed_audio.append(
                            Audio.from_base64(
                                aud_data["content"],
                                id=aud_data.get("id"),
                                mime_type=aud_data.get("mime_type"),
                                transcript=aud_data.get("transcript"),
                                expires_at=aud_data.get("expires_at"),
                                sample_rate=aud_data.get("sample_rate", 24000),
                                channels=aud_data.get("channels", 1),
                            )
                        )
                    else:
                        reconstructed_audio.append(Audio(**aud_data))
                else:
                    reconstructed_audio.append(aud_data)
            data["audio"] = reconstructed_audio

        # Handle video reconstruction properly
        if "videos" in data and data["videos"]:
            reconstructed_videos = []
            for i, vid_data in enumerate(data["videos"]):
                if isinstance(vid_data, dict):
                    # If content is base64, decode it back to bytes
                    if "content" in vid_data and isinstance(vid_data["content"], str):
                        reconstructed_videos.append(
                            Video.from_base64(
                                vid_data["content"],
                                id=vid_data.get("id"),
                                mime_type=vid_data.get("mime_type"),
                                format=vid_data.get("format"),
                            )
                        )
                    else:
                        reconstructed_videos.append(Video(**vid_data))
                else:
                    reconstructed_videos.append(vid_data)
            data["videos"] = reconstructed_videos

        # Handle file reconstruction properly
        if "files" in data and data["files"]:
            reconstructed_files = []
            for i, file_data in enumerate(data["files"]):
                if isinstance(file_data, dict):
                    # If content is base64, decode it back to bytes
                    if "content" in file_data and isinstance(file_data["content"], str):
                        reconstructed_files.append(
                            File.from_base64(
                                file_data["content"],
                                id=file_data.get("id"),
                                mime_type=file_data.get("mime_type"),
                                filename=file_data.get("filename"),
                                name=file_data.get("name"),
                                format=file_data.get("format"),
                            )
                        )
                    else:
                        reconstructed_files.append(File(**file_data))
                else:
                    reconstructed_files.append(file_data)
            data["files"] = reconstructed_files

        if "audio_output" in data and data["audio_output"]:
            aud_data = data["audio_output"]
            if isinstance(aud_data, dict):
                if "content" in aud_data and isinstance(aud_data["content"], str):
                    data["audio_output"] = Audio.from_base64(
                        aud_data["content"],
                        id=aud_data.get("id"),
                        mime_type=aud_data.get("mime_type"),
                        transcript=aud_data.get("transcript"),
                        expires_at=aud_data.get("expires_at"),
                        sample_rate=aud_data.get("sample_rate", 24000),
                        channels=aud_data.get("channels", 1),
                    )
                else:
                    data["audio_output"] = Audio(**aud_data)

        if "image_output" in data and data["image_output"]:
            img_data = data["image_output"]
            if isinstance(img_data, dict):
                if "content" in img_data and isinstance(img_data["content"], str):
                    data["image_output"] = Image.from_base64(
                        img_data["content"],
                        id=img_data.get("id"),
                        mime_type=img_data.get("mime_type"),
                        format=img_data.get("format"),
                    )
                else:
                    data["image_output"] = Image(**img_data)

        if "video_output" in data and data["video_output"]:
            vid_data = data["video_output"]
            if isinstance(vid_data, dict):
                if "content" in vid_data and isinstance(vid_data["content"], str):
                    data["video_output"] = Video.from_base64(
                        vid_data["content"],
                        id=vid_data.get("id"),
                        mime_type=vid_data.get("mime_type"),
                        format=vid_data.get("format"),
                    )
                else:
                    data["video_output"] = Video(**vid_data)

        return cls(**data)

    def to_dict(self) -> Dict[str, Any]:
        """Returns the message as a dictionary."""
        message_dict = {
            "id": self.id,
            "content": self.content,
            "reasoning_content": self.reasoning_content,
            "from_history": self.from_history,
            "stop_after_tool_call": self.stop_after_tool_call,
            "role": self.role,
            "name": self.name,
            "tool_call_id": self.tool_call_id,
            "tool_name": self.tool_name,
            "tool_args": self.tool_args,
            "tool_call_error": self.tool_call_error,
            "tool_calls": self.tool_calls,
            "redacted_reasoning_content": self.redacted_reasoning_content,
            "provider_data": self.provider_data,
        }
        # Filter out None and empty collections
        message_dict = {
            k: v for k, v in message_dict.items() if v is not None and not (isinstance(v, (list, dict)) and len(v) == 0)
        }

        # Convert media objects to dictionaries
        if self.images:
            message_dict["images"] = [img.to_dict() for img in self.images]
        if self.audio:
            message_dict["audio"] = [aud.to_dict() for aud in self.audio]
        if self.videos:
            message_dict["videos"] = [vid.to_dict() for vid in self.videos]
        if self.files:
            message_dict["files"] = [file.to_dict() for file in self.files]
        if self.audio_output:
            message_dict["audio_output"] = self.audio_output.to_dict()

        if self.references:
            message_dict["references"] = self.references.model_dump()
        if self.metrics:
            message_dict["metrics"] = self.metrics.to_dict()
            if not message_dict["metrics"]:
                message_dict.pop("metrics")

        message_dict["created_at"] = self.created_at
        return message_dict

    def to_function_call_dict(self) -> Dict[str, Any]:
        return {
            "content": self.content,
            "tool_call_id": self.tool_call_id,
            "tool_name": self.tool_name,
            "tool_args": self.tool_args,
            "tool_call_error": self.tool_call_error,
            "metrics": self.metrics,
            "created_at": self.created_at,
        }

    def log(self, metrics: bool = True, level: Optional[str] = None):
        """Log the message to the console

        Args:
            metrics (bool): Whether to log the metrics.
            level (str): The level to log the message at. One of debug, info, warning, or error.
                Defaults to debug.
        """
        _logger = log_debug
        if level == "info":
            _logger = log_info
        elif level == "warning":
            _logger = log_warning
        elif level == "error":
            _logger = log_error

        try:
            import shutil

            terminal_width = shutil.get_terminal_size().columns
        except Exception:
            terminal_width = 80  # fallback width

        header = f" {self.role} "
        _logger(f"{header.center(terminal_width - 20, '=')}")

        if self.name:
            _logger(f"Name: {self.name}")
        if self.tool_call_id:
            _logger(f"Tool call Id: {self.tool_call_id}")
        if self.reasoning_content:
            _logger(f"<reasoning>\n{self.reasoning_content}\n</reasoning>")
        if self.content:
            if isinstance(self.content, str) or isinstance(self.content, list):
                _logger(self.content)
            elif isinstance(self.content, dict):
                _logger(json.dumps(self.content, indent=2))
        if self.tool_calls:
            tool_calls_list = ["Tool Calls:"]
            for tool_call in self.tool_calls:
                tool_id = tool_call.get("id")
                function_name = tool_call.get("function", {}).get("name")
                tool_calls_list.append(f"  - ID: '{tool_id}'") if tool_id else None
                tool_calls_list.append(f"    Name: '{function_name}'") if function_name else None
                tool_call_arguments = tool_call.get("function", {}).get("arguments")
                if tool_call_arguments:
                    try:
                        tool_call_args: dict = (
                            tool_call_arguments
                            if isinstance(tool_call_arguments, dict)
                            else json.loads(tool_call_arguments)
                        )
                        if tool_call_args:
                            # Ensure tool_call_args is a dictionary before calling .items()
                            if isinstance(tool_call_args, dict):
                                arguments = ", ".join(f"{k}: {v}" for k, v in tool_call_args.items())
                                tool_calls_list.append(f"    Arguments: '{arguments}'")
                            else:
                                tool_calls_list.append(f"    Arguments: '{tool_call_args}'")
                    except json.JSONDecodeError:
                        tool_calls_list.append("    Arguments: 'Invalid JSON format'")
            tool_calls_str = "\n".join(tool_calls_list)

            _logger(tool_calls_str)
        if self.images:
            _logger(f"Images added: {len(self.images)}")
        if self.videos:
            _logger(f"Videos added: {len(self.videos)}")
        if self.audio:
            _logger(f"Audio Files added: {len(self.audio)}")
        if self.files:
            _logger(f"Files added: {len(self.files)}")

        metrics_header = " TOOL METRICS " if self.role == "tool" else " METRICS "
        if metrics and self.metrics is not None and self.metrics != Metrics():
            _logger(metrics_header, center=True, symbol="*")

            # Token metrics
            token_metrics = []
            if self.metrics.input_tokens and self.metrics.input_tokens > 0:
                token_metrics.append(f"input={self.metrics.input_tokens}")
            if self.metrics.output_tokens and self.metrics.output_tokens > 0:
                token_metrics.append(f"output={self.metrics.output_tokens}")
            if self.metrics.total_tokens and self.metrics.total_tokens > 0:
                token_metrics.append(f"total={self.metrics.total_tokens}")
            if self.metrics.cache_read_tokens and self.metrics.cache_read_tokens > 0:
                token_metrics.append(f"cached={self.metrics.cache_read_tokens}")
            if self.metrics.cache_write_tokens and self.metrics.cache_write_tokens > 0:
                token_metrics.append(f"cache_write_tokens={self.metrics.cache_write_tokens}")
            if self.metrics.reasoning_tokens and self.metrics.reasoning_tokens > 0:
                token_metrics.append(f"reasoning={self.metrics.reasoning_tokens}")
            if self.metrics.audio_total_tokens and self.metrics.audio_total_tokens > 0:
                token_metrics.append(f"audio={self.metrics.audio_total_tokens}")
            if token_metrics:
                _logger(f"* Tokens:                      {', '.join(token_metrics)}")

            # Time related metrics
            if self.metrics.duration is not None and self.metrics.duration > 0:
                _logger(f"* Duration:                    {self.metrics.duration:.4f}s")
            if self.metrics.output_tokens and self.metrics.duration and self.metrics.duration > 0:
                _logger(
                    f"* Tokens per second:           {self.metrics.output_tokens / self.metrics.duration:.4f} tokens/s"
                )
            if self.metrics.time_to_first_token is not None and self.metrics.time_to_first_token > 0:
                _logger(f"* Time to first token:         {self.metrics.time_to_first_token:.4f}s")

            # Non-generic metrics
            if self.metrics.provider_metrics:
                _logger(f"* Provider metrics:            {self.metrics.provider_metrics}")
            if self.metrics.additional_metrics:
                _logger(f"* Additional metrics:          {self.metrics.additional_metrics}")

            _logger(metrics_header, center=True, symbol="*")

    def content_is_valid(self) -> bool:
        """Check if the message content is valid."""
        return self.content is not None and len(self.content) > 0



from dataclasses import asdict, dataclass, field
from enum import Enum
from time import time
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from pydantic import BaseModel

from agno.media import Audio, Image, Video
from agno.run.agent import RunOutput
from agno.run.base import BaseRunOutputEvent, RunStatus
from agno.run.team import TeamRunOutput

if TYPE_CHECKING:
    from agno.workflow.types import StepOutput, WorkflowMetrics
else:
    StepOutput = Any
    WorkflowMetrics = Any


class WorkflowRunEvent(str, Enum):
    """Events that can be sent by workflow execution"""

    workflow_started = "WorkflowStarted"
    workflow_completed = "WorkflowCompleted"
    workflow_cancelled = "WorkflowCancelled"
    workflow_error = "WorkflowError"

    step_started = "StepStarted"
    step_completed = "StepCompleted"
    step_error = "StepError"

    loop_execution_started = "LoopExecutionStarted"
    loop_iteration_started = "LoopIterationStarted"
    loop_iteration_completed = "LoopIterationCompleted"
    loop_execution_completed = "LoopExecutionCompleted"

    parallel_execution_started = "ParallelExecutionStarted"
    parallel_execution_completed = "ParallelExecutionCompleted"

    condition_execution_started = "ConditionExecutionStarted"
    condition_execution_completed = "ConditionExecutionCompleted"

    router_execution_started = "RouterExecutionStarted"
    router_execution_completed = "RouterExecutionCompleted"

    steps_execution_started = "StepsExecutionStarted"
    steps_execution_completed = "StepsExecutionCompleted"

    step_output = "StepOutput"

    custom_event = "CustomEvent"


@dataclass
class BaseWorkflowRunOutputEvent(BaseRunOutputEvent):
    """Base class for all workflow run response events"""

    created_at: int = field(default_factory=lambda: int(time()))
    event: str = ""

    # Workflow-specific fields
    workflow_id: Optional[str] = None
    workflow_name: Optional[str] = None
    session_id: Optional[str] = None
    run_id: Optional[str] = None
    step_id: Optional[str] = None
    parent_step_id: Optional[str] = None

    def to_dict(self) -> Dict[str, Any]:
        _dict = {k: v for k, v in asdict(self).items() if v is not None}

        if hasattr(self, "content") and self.content and isinstance(self.content, BaseModel):
            _dict["content"] = self.content.model_dump(exclude_none=True)

        # Handle StepOutput fields that contain Message objects
        if hasattr(self, "step_results") and self.step_results is not None:
            _dict["step_results"] = [step.to_dict() if hasattr(step, "to_dict") else step for step in self.step_results]

        if hasattr(self, "step_response") and self.step_response is not None:
            _dict["step_response"] = (
                self.step_response.to_dict() if hasattr(self.step_response, "to_dict") else self.step_response
            )

        if hasattr(self, "iteration_results") and self.iteration_results is not None:
            _dict["iteration_results"] = [
                step.to_dict() if hasattr(step, "to_dict") else step for step in self.iteration_results
            ]

        if hasattr(self, "all_results") and self.all_results is not None:
            _dict["all_results"] = [
                [step.to_dict() if hasattr(step, "to_dict") else step for step in iteration]
                for iteration in self.all_results
            ]

        return _dict

    @property
    def is_cancelled(self):
        return False

    @property
    def is_error(self):
        return False

    @property
    def status(self):
        status = "Completed"
        if self.is_error:
            status = "Error"
        if self.is_cancelled:
            status = "Cancelled"

        return status


@dataclass
class WorkflowStartedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when workflow execution starts"""

    event: str = WorkflowRunEvent.workflow_started.value


@dataclass
class WorkflowCompletedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when workflow execution completes"""

    event: str = WorkflowRunEvent.workflow_completed.value
    content: Optional[Any] = None
    content_type: str = "str"

    # Store actual step execution results as StepOutput objects
    step_results: List[StepOutput] = field(default_factory=list)
    metadata: Optional[Dict[str, Any]] = None


@dataclass
class WorkflowErrorEvent(BaseWorkflowRunOutputEvent):
    """Event sent when workflow execution fails"""

    event: str = WorkflowRunEvent.workflow_error.value
    error: Optional[str] = None

    # From exceptions
    error_type: Optional[str] = None
    error_id: Optional[str] = None
    additional_data: Optional[Dict[str, Any]] = None


@dataclass
class WorkflowCancelledEvent(BaseWorkflowRunOutputEvent):
    """Event sent when workflow execution is cancelled"""

    event: str = WorkflowRunEvent.workflow_cancelled.value
    reason: Optional[str] = None

    @property
    def is_cancelled(self):
        return True


@dataclass
class StepStartedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when step execution starts"""

    event: str = WorkflowRunEvent.step_started.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None


@dataclass
class StepCompletedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when step execution completes"""

    event: str = WorkflowRunEvent.step_completed.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None

    content: Optional[Any] = None
    content_type: str = "str"

    # Media content fields
    images: Optional[List[Image]] = None
    videos: Optional[List[Video]] = None
    audio: Optional[List[Audio]] = None
    response_audio: Optional[Audio] = None

    # Store actual step execution results as StepOutput objects
    step_response: Optional[StepOutput] = None


@dataclass
class StepErrorEvent(BaseWorkflowRunOutputEvent):
    """Event sent when step execution fails"""

    event: str = WorkflowRunEvent.step_error.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    error: Optional[str] = None


@dataclass
class LoopExecutionStartedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when loop execution starts"""

    event: str = WorkflowRunEvent.loop_execution_started.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    max_iterations: Optional[int] = None


@dataclass
class LoopIterationStartedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when loop iteration starts"""

    event: str = WorkflowRunEvent.loop_iteration_started.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    iteration: int = 0
    max_iterations: Optional[int] = None


@dataclass
class LoopIterationCompletedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when loop iteration completes"""

    event: str = WorkflowRunEvent.loop_iteration_completed.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    iteration: int = 0
    max_iterations: Optional[int] = None
    iteration_results: List[StepOutput] = field(default_factory=list)
    should_continue: bool = True


@dataclass
class LoopExecutionCompletedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when loop execution completes"""

    event: str = WorkflowRunEvent.loop_execution_completed.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    total_iterations: int = 0
    max_iterations: Optional[int] = None
    all_results: List[List[StepOutput]] = field(default_factory=list)


@dataclass
class ParallelExecutionStartedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when parallel step execution starts"""

    event: str = WorkflowRunEvent.parallel_execution_started.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    parallel_step_count: Optional[int] = None


@dataclass
class ParallelExecutionCompletedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when parallel step execution completes"""

    event: str = WorkflowRunEvent.parallel_execution_completed.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    parallel_step_count: Optional[int] = None

    # Results from all parallel steps
    step_results: List[StepOutput] = field(default_factory=list)


@dataclass
class ConditionExecutionStartedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when condition step execution starts"""

    event: str = WorkflowRunEvent.condition_execution_started.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    condition_result: Optional[bool] = None


@dataclass
class ConditionExecutionCompletedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when condition step execution completes"""

    event: str = WorkflowRunEvent.condition_execution_completed.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    condition_result: Optional[bool] = None
    executed_steps: Optional[int] = None

    # Results from executed steps
    step_results: List[StepOutput] = field(default_factory=list)


@dataclass
class RouterExecutionStartedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when router step execution starts"""

    event: str = WorkflowRunEvent.router_execution_started.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    # Names of steps selected by router
    selected_steps: List[str] = field(default_factory=list)


@dataclass
class RouterExecutionCompletedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when router step execution completes"""

    event: str = WorkflowRunEvent.router_execution_completed.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    # Names of steps that were selected
    selected_steps: List[str] = field(default_factory=list)
    executed_steps: Optional[int] = None

    # Results from executed steps
    step_results: List[StepOutput] = field(default_factory=list)


@dataclass
class StepsExecutionStartedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when steps execution starts"""

    event: str = WorkflowRunEvent.steps_execution_started.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    steps_count: Optional[int] = None


@dataclass
class StepsExecutionCompletedEvent(BaseWorkflowRunOutputEvent):
    """Event sent when steps execution completes"""

    event: str = WorkflowRunEvent.steps_execution_completed.value
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None
    steps_count: Optional[int] = None
    executed_steps: Optional[int] = None

    # Results from executed steps
    step_results: List[StepOutput] = field(default_factory=list)


@dataclass
class StepOutputEvent(BaseWorkflowRunOutputEvent):
    """Event sent when a step produces output - replaces direct StepOutput yielding"""

    event: str = "StepOutput"
    step_name: Optional[str] = None
    step_index: Optional[Union[int, tuple]] = None

    # Store actual step execution result as StepOutput object
    step_output: Optional[StepOutput] = None

    # Properties for backward compatibility
    @property
    def content(self) -> Optional[Union[str, Dict[str, Any], List[Any], BaseModel, Any]]:
        return self.step_output.content if self.step_output else None

    @property
    def images(self) -> Optional[List[Image]]:
        return self.step_output.images if self.step_output else None

    @property
    def videos(self) -> Optional[List[Video]]:
        return self.step_output.videos if self.step_output else None

    @property
    def audio(self) -> Optional[List[Audio]]:
        return self.step_output.audio if self.step_output else None

    @property
    def success(self) -> bool:
        return self.step_output.success if self.step_output else True

    @property
    def error(self) -> Optional[str]:
        return self.step_output.error if self.step_output else None

    @property
    def stop(self) -> bool:
        return self.step_output.stop if self.step_output else False


@dataclass
class CustomEvent(BaseWorkflowRunOutputEvent):
    """Event sent when a custom event is produced"""

    event: str = WorkflowRunEvent.custom_event.value


# Union type for all workflow run response events
WorkflowRunOutputEvent = Union[
    WorkflowStartedEvent,
    WorkflowCompletedEvent,
    WorkflowErrorEvent,
    WorkflowCancelledEvent,
    StepStartedEvent,
    StepCompletedEvent,
    StepErrorEvent,
    LoopExecutionStartedEvent,
    LoopIterationStartedEvent,
    LoopIterationCompletedEvent,
    LoopExecutionCompletedEvent,
    ParallelExecutionStartedEvent,
    ParallelExecutionCompletedEvent,
    ConditionExecutionStartedEvent,
    ConditionExecutionCompletedEvent,
    RouterExecutionStartedEvent,
    RouterExecutionCompletedEvent,
    StepsExecutionStartedEvent,
    StepsExecutionCompletedEvent,
    StepOutputEvent,
    CustomEvent,
]

# Map event string to dataclass for workflow events
WORKFLOW_RUN_EVENT_TYPE_REGISTRY = {
    WorkflowRunEvent.workflow_started.value: WorkflowStartedEvent,
    WorkflowRunEvent.workflow_completed.value: WorkflowCompletedEvent,
    WorkflowRunEvent.workflow_cancelled.value: WorkflowCancelledEvent,
    WorkflowRunEvent.workflow_error.value: WorkflowErrorEvent,
    WorkflowRunEvent.step_started.value: StepStartedEvent,
    WorkflowRunEvent.step_completed.value: StepCompletedEvent,
    WorkflowRunEvent.step_error.value: StepErrorEvent,
    WorkflowRunEvent.loop_execution_started.value: LoopExecutionStartedEvent,
    WorkflowRunEvent.loop_iteration_started.value: LoopIterationStartedEvent,
    WorkflowRunEvent.loop_iteration_completed.value: LoopIterationCompletedEvent,
    WorkflowRunEvent.loop_execution_completed.value: LoopExecutionCompletedEvent,
    WorkflowRunEvent.parallel_execution_started.value: ParallelExecutionStartedEvent,
    WorkflowRunEvent.parallel_execution_completed.value: ParallelExecutionCompletedEvent,
    WorkflowRunEvent.condition_execution_started.value: ConditionExecutionStartedEvent,
    WorkflowRunEvent.condition_execution_completed.value: ConditionExecutionCompletedEvent,
    WorkflowRunEvent.router_execution_started.value: RouterExecutionStartedEvent,
    WorkflowRunEvent.router_execution_completed.value: RouterExecutionCompletedEvent,
    WorkflowRunEvent.steps_execution_started.value: StepsExecutionStartedEvent,
    WorkflowRunEvent.steps_execution_completed.value: StepsExecutionCompletedEvent,
    WorkflowRunEvent.step_output.value: StepOutputEvent,
    WorkflowRunEvent.custom_event.value: CustomEvent,
}


def workflow_run_output_event_from_dict(data: dict) -> BaseWorkflowRunOutputEvent:
    event_type = data.get("event", "")
    cls = WORKFLOW_RUN_EVENT_TYPE_REGISTRY.get(event_type)
    if not cls:
        raise ValueError(f"Unknown workflow event type: {event_type}")
    return cls.from_dict(data)  # type: ignore


@dataclass
class WorkflowRunOutput:
    """Response returned by Workflow.run() functions - kept for backwards compatibility"""

    input: Optional[Union[str, Dict[str, Any], List[Any], BaseModel]] = None
    content: Optional[Union[str, Dict[str, Any], List[Any], BaseModel, Any]] = None
    content_type: str = "str"

    # Workflow-specific fields
    workflow_id: Optional[str] = None
    workflow_name: Optional[str] = None

    run_id: Optional[str] = None
    session_id: Optional[str] = None

    # Media content fields
    images: Optional[List[Image]] = None
    videos: Optional[List[Video]] = None
    audio: Optional[List[Audio]] = None
    response_audio: Optional[Audio] = None

    # Store actual step execution results as StepOutput objects
    step_results: List[Union[StepOutput, List[StepOutput]]] = field(default_factory=list)

    # Store agent/team responses separately with parent_run_id references
    step_executor_runs: Optional[List[Union[RunOutput, TeamRunOutput]]] = None

    # Store events from workflow execution
    events: Optional[List[WorkflowRunOutputEvent]] = None

    # Workflow metrics aggregated from all steps
    metrics: Optional[WorkflowMetrics] = None

    metadata: Optional[Dict[str, Any]] = None
    created_at: int = field(default_factory=lambda: int(time()))

    status: RunStatus = RunStatus.pending

    @property
    def is_cancelled(self):
        return self.status == RunStatus.cancelled

    def to_dict(self) -> Dict[str, Any]:
        _dict = {
            k: v
            for k, v in asdict(self).items()
            if v is not None
            and k
            not in [
                "metadata",
                "images",
                "videos",
                "audio",
                "response_audio",
                "step_results",
                "step_executor_runs",
                "events",
                "metrics",
            ]
        }

        if self.status is not None:
            _dict["status"] = self.status.value if isinstance(self.status, RunStatus) else self.status

        if self.metadata is not None:
            _dict["metadata"] = self.metadata

        if self.images is not None:
            _dict["images"] = [img.to_dict() for img in self.images]

        if self.videos is not None:
            _dict["videos"] = [vid.to_dict() for vid in self.videos]

        if self.audio is not None:
            _dict["audio"] = [aud.to_dict() for aud in self.audio]

        if self.response_audio is not None:
            _dict["response_audio"] = self.response_audio.to_dict()

        if self.step_results:
            flattened_responses = []
            for step_response in self.step_results:
                if isinstance(step_response, list):
                    # Handle List[StepOutput] from workflow components like Steps
                    flattened_responses.extend([s.to_dict() for s in step_response])
                else:
                    # Handle single StepOutput
                    flattened_responses.append(step_response.to_dict())
            _dict["step_results"] = flattened_responses

        if self.step_executor_runs:
            _dict["step_executor_runs"] = [run.to_dict() for run in self.step_executor_runs]

        if self.metrics is not None:
            _dict["metrics"] = self.metrics.to_dict()

        if self.input is not None:
            if isinstance(self.input, BaseModel):
                _dict["input"] = self.input.model_dump(exclude_none=True)
            else:
                _dict["input"] = self.input

        if self.content and isinstance(self.content, BaseModel):
            _dict["content"] = self.content.model_dump(exclude_none=True)

        if self.events is not None:
            _dict["events"] = [e.to_dict() for e in self.events]

        return _dict

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "WorkflowRunOutput":
        # Import here to avoid circular import
        from agno.workflow.step import StepOutput

        workflow_metrics_dict = data.pop("metrics", {})
        workflow_metrics = None
        if workflow_metrics_dict:
            from agno.workflow.workflow import WorkflowMetrics

            workflow_metrics = WorkflowMetrics.from_dict(workflow_metrics_dict)

        step_results = data.pop("step_results", [])
        parsed_step_results: List[Union[StepOutput, List[StepOutput]]] = []
        if step_results:
            for step_output_dict in step_results:
                # Reconstruct StepOutput from dict
                parsed_step_results.append(StepOutput.from_dict(step_output_dict))

        # Parse step_executor_runs
        step_executor_runs_data = data.pop("step_executor_runs", [])
        step_executor_runs: List[Union[RunOutput, TeamRunOutput]] = []
        if step_executor_runs_data:
            step_executor_runs = []
            for run_data in step_executor_runs_data:
                if "team_id" in run_data or "team_name" in run_data:
                    step_executor_runs.append(TeamRunOutput.from_dict(run_data))
                else:
                    step_executor_runs.append(RunOutput.from_dict(run_data))

        metadata = data.pop("metadata", None)

        images = data.pop("images", [])
        images = [Image.model_validate(image) for image in images] if images else None

        videos = data.pop("videos", [])
        videos = [Video.model_validate(video) for video in videos] if videos else None

        audio = data.pop("audio", [])
        audio = [Audio.model_validate(audio) for audio in audio] if audio else None

        response_audio = data.pop("response_audio", None)
        response_audio = Audio.model_validate(response_audio) if response_audio else None

        events_data = data.pop("events", [])
        final_events = []
        for event in events_data or []:
            if "agent_id" in event:
                # Agent event from agent step
                from agno.run.agent import run_output_event_from_dict

                event = run_output_event_from_dict(event)
            elif "team_id" in event:
                # Team event from team step
                from agno.run.team import team_run_output_event_from_dict

                event = team_run_output_event_from_dict(event)
            else:
                # Pure workflow event
                event = workflow_run_output_event_from_dict(event)
            final_events.append(event)
        events = final_events

        input_data = data.pop("input", None)

        return cls(
            step_results=parsed_step_results,
            metadata=metadata,
            images=images,
            videos=videos,
            audio=audio,
            response_audio=response_audio,
            events=events,
            metrics=workflow_metrics,
            step_executor_runs=step_executor_runs,
            input=input_data,
            **data,
        )

    def get_content_as_string(self, **kwargs) -> str:
        import json

        from pydantic import BaseModel

        if isinstance(self.content, str):
            return self.content
        elif isinstance(self.content, BaseModel):
            return self.content.model_dump_json(exclude_none=True, **kwargs)
        else:
            return json.dumps(self.content, **kwargs)

    def has_completed(self) -> bool:
        """Check if the workflow run is completed (either successfully or with error)"""
        return self.status in [RunStatus.completed, RunStatus.error]



"""Utility functions for the MongoDB database class."""

import json
import time
from datetime import date, datetime, timedelta, timezone
from typing import Any, Dict, List, Optional
from uuid import uuid4

from agno.db.mongo.schemas import get_collection_indexes
from agno.utils.log import log_error, log_warning

try:
    from pymongo.collection import Collection
except ImportError:
    raise ImportError("`pymongo` not installed. Please install it using `pip install pymongo`")


# -- DB util methods --
def create_collection_indexes(collection: Collection, collection_type: str) -> None:
    """Create all required indexes for a collection"""
    try:
        indexes = get_collection_indexes(collection_type)
        for index_spec in indexes:
            key = index_spec["key"]
            unique = index_spec.get("unique", False)

            if isinstance(key, list):
                collection.create_index(key, unique=unique)
            else:
                collection.create_index([(key, 1)], unique=unique)

    except Exception as e:
        log_warning(f"Error creating indexes for {collection_type} collection: {e}")


def apply_sorting(
    query_args: Dict[str, Any], sort_by: Optional[str] = None, sort_order: Optional[str] = None
) -> List[tuple]:
    """Apply sorting to MongoDB query."""
    if sort_by is None:
        return []

    sort_direction = 1 if sort_order == "asc" else -1
    return [(sort_by, sort_direction)]


def apply_pagination(
    query_args: Dict[str, Any], limit: Optional[int] = None, page: Optional[int] = None
) -> Dict[str, Any]:
    """Apply pagination to MongoDB query."""
    if limit is not None:
        query_args["limit"] = limit
        if page is not None:
            query_args["skip"] = (page - 1) * limit
    return query_args


# -- Metrics util methods --
def calculate_date_metrics(date_to_process: date, sessions_data: dict) -> dict:
    """Calculate metrics for the given single date."""
    metrics = {
        "users_count": 0,
        "agent_sessions_count": 0,
        "team_sessions_count": 0,
        "workflow_sessions_count": 0,
        "agent_runs_count": 0,
        "team_runs_count": 0,
        "workflow_runs_count": 0,
    }
    token_metrics = {
        "input_tokens": 0,
        "output_tokens": 0,
        "total_tokens": 0,
        "audio_total_tokens": 0,
        "audio_input_tokens": 0,
        "audio_output_tokens": 0,
        "cache_read_tokens": 0,
        "cache_write_tokens": 0,
        "reasoning_tokens": 0,
    }
    model_counts: Dict[str, int] = {}

    session_types = [
        ("agent", "agent_sessions_count", "agent_runs_count"),
        ("team", "team_sessions_count", "team_runs_count"),
        ("workflow", "workflow_sessions_count", "workflow_runs_count"),
    ]
    all_user_ids = set()

    for session_type, sessions_count_key, runs_count_key in session_types:
        sessions = sessions_data.get(session_type, [])
        metrics[sessions_count_key] = len(sessions)

        for session in sessions:
            if session.get("user_id"):
                all_user_ids.add(session["user_id"])
            runs = session.get("runs", []) or []
            metrics[runs_count_key] += len(sessions)

            if runs := session.get("runs", []):
                if isinstance(runs, str):
                    runs = json.loads(runs)
                for run in runs:
                    if model_id := run.get("model"):
                        model_provider = run.get("model_provider", "")
                        model_counts[f"{model_id}:{model_provider}"] = (
                            model_counts.get(f"{model_id}:{model_provider}", 0) + 1
                        )

            session_data = session.get("session_data", {})
            if isinstance(session_data, str):
                session_data = json.loads(session_data)
            session_metrics = session_data.get("session_metrics", {})
            for field in token_metrics:
                token_metrics[field] += session_metrics.get(field, 0)

    model_metrics = []
    for model, count in model_counts.items():
        model_id, model_provider = model.split(":")
        model_metrics.append({"model_id": model_id, "model_provider": model_provider, "count": count})

    metrics["users_count"] = len(all_user_ids)
    current_time = int(time.time())

    return {
        "id": str(uuid4()),
        "date": date_to_process,
        "completed": date_to_process < datetime.now(timezone.utc).date(),
        "token_metrics": token_metrics,
        "model_metrics": model_metrics,
        "created_at": current_time,
        "updated_at": current_time,
        "aggregation_period": "daily",
        **metrics,
    }


def fetch_all_sessions_data(
    sessions: List[Dict[str, Any]], dates_to_process: list[date], start_timestamp: int
) -> Optional[dict]:
    """Return all session data for the given dates, for all session types."""
    if not dates_to_process:
        return None

    all_sessions_data: Dict[str, Dict[str, List[Dict[str, Any]]]] = {
        date_to_process.isoformat(): {"agent": [], "team": [], "workflow": []} for date_to_process in dates_to_process
    }

    for session in sessions:
        session_date = (
            datetime.fromtimestamp(session.get("created_at", start_timestamp), tz=timezone.utc).date().isoformat()
        )
        if session_date in all_sessions_data:
            all_sessions_data[session_date][session["session_type"]].append(session)

    return all_sessions_data


def get_dates_to_calculate_metrics_for(starting_date: date) -> list[date]:
    """Return the list of dates to calculate metrics for."""
    today = datetime.now(timezone.utc).date()
    days_diff = (today - starting_date).days + 1
    if days_diff <= 0:
        return []
    return [starting_date + timedelta(days=x) for x in range(days_diff)]


def bulk_upsert_metrics(collection: Collection, metrics_records: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """Bulk upsert metrics into the database.

    Args:
        collection (Collection): The collection to upsert the metrics into.
        metrics_records (List[Dict[str, Any]]): The list of metrics records to upsert.

    Returns:
        The list of upserted metrics records.
    """
    if not metrics_records:
        return []

    results = []
    for record in metrics_records:
        record["date"] = record["date"].isoformat() if isinstance(record["date"], date) else record["date"]
        try:
            collection.replace_one(
                {
                    "date": record["date"],
                    "aggregation_period": record["aggregation_period"],
                },
                record,
                upsert=True,
            )

            results.append(record)

        except Exception as e:
            log_error(f"Error upserting metrics record: {e}")
            continue

    return results



import json

try:
    from spider import Spider as ExternalSpider
except ImportError:
    raise ImportError("`spider-client` not installed. Please install using `pip install spider-client`")

from typing import Any, List, Optional

from agno.tools import Toolkit
from agno.utils.log import log_info, logger


class SpiderTools(Toolkit):
    """
    Spider is a toolkit for web searching, scraping, and crawling.

    Args:
        enable_search (bool): Enable web search functionality. Default is True.
        enable_scrape (bool): Enable web scraping functionality. Default is True.
        enable_crawl (bool): Enable web crawling functionality. Default is True.
        all (bool): Enable all tools. Overrides individual flags when True. Default is False.
        max_results (Optional[int]): Default maximum number of results.
        url (Optional[str]): Default URL for operations.
        optional_params (Optional[dict]): Additional parameters for operations.
    """

    def __init__(
        self,
        max_results: Optional[int] = None,
        url: Optional[str] = None,
        optional_params: Optional[dict] = None,
        enable_search: bool = True,
        enable_scrape: bool = True,
        enable_crawl: bool = True,
        all: bool = False,
        **kwargs,
    ):
        self.max_results = max_results
        self.url = url
        self.optional_params = optional_params or {}

        tools: List[Any] = []
        if enable_search or all:
            tools.append(self.search_web)
        if enable_scrape or all:
            tools.append(self.scrape)
        if enable_crawl or all:
            tools.append(self.crawl)

        super().__init__(name="spider", tools=tools, **kwargs)

    def search_web(self, query: str, max_results: int = 5) -> str:
        """Use this function to search the web.
        Args:
            query (str): The query to search the web with.
            max_results (int, optional): The maximum number of results to return. Defaults to 5.
        Returns:
            The results of the search.
        """
        max_results = self.max_results or max_results
        return self._search(query, max_results=max_results)

    def scrape(self, url: str) -> str:
        """Use this function to scrape the content of a webpage.
        Args:
            url (str): The URL of the webpage to scrape.
        Returns:
            Markdown of the webpage.
        """
        return self._scrape(url)

    def crawl(self, url: str, limit: Optional[int] = None) -> str:
        """Use this function to crawl the web.
        Args:
            url (str): The URL of the webpage to crawl.
            limit (int, optional): The maximum number of pages to crawl. Defaults to 10.
        Returns:
            The results of the crawl.
        """
        return self._crawl(url, limit=limit)

    def _search(self, query: str, max_results: int = 1) -> str:
        app = ExternalSpider()
        log_info(f"Fetching results from spider for query: {query} with max_results: {max_results}")
        try:
            options = {"fetch_page_content": False, "num": max_results, **self.optional_params}
            results = app.search(query, options)
            return json.dumps(results)
        except Exception as e:
            logger.error(f"Error fetching results from spider: {e}")
            return f"Error fetching results from spider: {e}"

    def _scrape(self, url: str) -> str:
        app = ExternalSpider()
        log_info(f"Fetching content from spider for url: {url}")
        try:
            options = {"return_format": "markdown", **self.optional_params}
            results = app.scrape_url(url, options)
            return json.dumps(results)
        except Exception as e:
            logger.error(f"Error fetching content from spider: {e}")
            return f"Error fetching content from spider: {e}"

    def _crawl(self, url: str, limit: Optional[int] = None) -> str:
        app = ExternalSpider()
        log_info(f"Fetching content from spider for url: {url}")
        try:
            if limit is None:
                limit = 10
            options = {"return_format": "markdown", "limit": limit, **self.optional_params}
            results = app.crawl_url(url, options)
            return json.dumps(results)
        except Exception as e:
            logger.error(f"Error fetching content from spider: {e}")
            return f"Error fetching content from spider: {e}"



import json
from collections.abc import AsyncIterator
from dataclasses import asdict, dataclass
from os import getenv
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseModel

from agno.exceptions import ModelProviderError, ModelRateLimitError
from agno.models.base import Model
from agno.models.message import Citations, DocumentCitation, Message, UrlCitation
from agno.models.metrics import Metrics
from agno.models.response import ModelResponse
from agno.run.agent import RunOutput
from agno.utils.log import log_debug, log_error, log_warning
from agno.utils.models.claude import MCPServerConfiguration, format_messages, format_tools_for_model

try:
    from anthropic import Anthropic as AnthropicClient
    from anthropic import (
        APIConnectionError,
        APIStatusError,
        RateLimitError,
    )
    from anthropic import (
        AsyncAnthropic as AsyncAnthropicClient,
    )
    from anthropic.types import (
        CitationPageLocation,
        CitationsWebSearchResultLocation,
        ContentBlockDeltaEvent,
        ContentBlockStartEvent,
        ContentBlockStopEvent,
        MessageDeltaUsage,
        # MessageDeltaEvent,  # Currently broken
        MessageStopEvent,
        Usage,
    )
    from anthropic.types import (
        Message as AnthropicMessage,
    )
except ImportError as e:
    raise ImportError("`anthropic` not installed. Please install it with `pip install anthropic`") from e

# Import Beta types
try:
    from anthropic.types.beta import BetaRawContentBlockDeltaEvent, BetaTextDelta
except ImportError as e:
    raise ImportError(
        "`anthropic` not installed or missing beta components. Please install with `pip install anthropic`"
    ) from e


@dataclass
class Claude(Model):
    """
    A class representing Anthropic Claude model.

    For more information, see: https://docs.anthropic.com/en/api/messages
    """

    # Models that DO NOT support extended thinking
    # All future models are assumed to support thinking
    # Based on official Anthropic documentation: https://docs.claude.com/en/docs/about-claude/models/overview
    NON_THINKING_MODELS = {
        # Claude Haiku 3 family (does not support thinking)
        "claude-3-haiku-20240307",
        # Claude Haiku 3.5 family (does not support thinking)
        "claude-3-5-haiku-20241022",
        "claude-3-5-haiku-latest",
    }

    id: str = "claude-sonnet-4-5-20250929"
    name: str = "Claude"
    provider: str = "Anthropic"

    # Request parameters
    max_tokens: Optional[int] = 4096
    thinking: Optional[Dict[str, Any]] = None
    temperature: Optional[float] = None
    stop_sequences: Optional[List[str]] = None
    top_p: Optional[float] = None
    top_k: Optional[int] = None
    cache_system_prompt: Optional[bool] = False
    extended_cache_time: Optional[bool] = False
    request_params: Optional[Dict[str, Any]] = None
    mcp_servers: Optional[List[MCPServerConfiguration]] = None

    # Client parameters
    api_key: Optional[str] = None
    default_headers: Optional[Dict[str, Any]] = None
    timeout: Optional[float] = None
    client_params: Optional[Dict[str, Any]] = None

    # Anthropic clients
    client: Optional[AnthropicClient] = None
    async_client: Optional[AsyncAnthropicClient] = None

    def __post_init__(self):
        """Validate model configuration after initialization"""
        # Validate thinking support immediately at model creation
        if self.thinking:
            self._validate_thinking_support()

    def _get_client_params(self) -> Dict[str, Any]:
        client_params: Dict[str, Any] = {}

        self.api_key = self.api_key or getenv("ANTHROPIC_API_KEY")
        if not self.api_key:
            log_error("ANTHROPIC_API_KEY not set. Please set the ANTHROPIC_API_KEY environment variable.")

        # Add API key to client parameters
        client_params["api_key"] = self.api_key
        if self.timeout is not None:
            client_params["timeout"] = self.timeout

        # Add additional client parameters
        if self.client_params is not None:
            client_params.update(self.client_params)
        if self.default_headers is not None:
            client_params["default_headers"] = self.default_headers
        return client_params

    def get_client(self) -> AnthropicClient:
        """
        Returns an instance of the Anthropic client.
        """
        if self.client and not self.client.is_closed():
            return self.client

        _client_params = self._get_client_params()
        self.client = AnthropicClient(**_client_params)
        return self.client

    def get_async_client(self) -> AsyncAnthropicClient:
        """
        Returns an instance of the async Anthropic client.
        """
        if self.async_client:
            return self.async_client

        _client_params = self._get_client_params()
        self.async_client = AsyncAnthropicClient(**_client_params)
        return self.async_client

    def _validate_thinking_support(self) -> None:
        """
        Validate that the current model supports extended thinking.

        Raises:
            ValueError: If thinking is enabled but the model doesn't support it
        """
        if self.thinking and self.id in self.NON_THINKING_MODELS:
            non_thinking_models = "\n  - ".join(sorted(self.NON_THINKING_MODELS))
            raise ValueError(
                f"Model '{self.id}' does not support extended thinking.\n\n"
                f"The following models do NOT support thinking:\n  - {non_thinking_models}\n\n"
                f"All other Claude models support extended thinking by default.\n"
                f"For more information, see: https://docs.anthropic.com/en/docs/about-claude/models/overview"
            )

    def get_request_params(self) -> Dict[str, Any]:
        """
        Generate keyword arguments for API requests.
        """
        # Validate thinking support if thinking is enabled
        if self.thinking:
            self._validate_thinking_support()

        _request_params: Dict[str, Any] = {}
        if self.max_tokens:
            _request_params["max_tokens"] = self.max_tokens
        if self.thinking:
            _request_params["thinking"] = self.thinking
        if self.temperature:
            _request_params["temperature"] = self.temperature
        if self.stop_sequences:
            _request_params["stop_sequences"] = self.stop_sequences
        if self.top_p:
            _request_params["top_p"] = self.top_p
        if self.top_k:
            _request_params["top_k"] = self.top_k
        if self.mcp_servers:
            _request_params["mcp_servers"] = [
                {k: v for k, v in asdict(server).items() if v is not None} for server in self.mcp_servers
            ]
        if self.request_params:
            _request_params.update(self.request_params)

        return _request_params

    def _prepare_request_kwargs(
        self, system_message: str, tools: Optional[List[Dict[str, Any]]] = None
    ) -> Dict[str, Any]:
        """
        Prepare the request keyword arguments for the API call.

        Args:
            system_message (str): The concatenated system messages.

        Returns:
            Dict[str, Any]: The request keyword arguments.
        """
        request_kwargs = self.get_request_params().copy()
        if system_message:
            if self.cache_system_prompt:
                cache_control = (
                    {"type": "ephemeral", "ttl": "1h"}
                    if self.extended_cache_time is not None and self.extended_cache_time is True
                    else {"type": "ephemeral"}
                )
                request_kwargs["system"] = [{"text": system_message, "type": "text", "cache_control": cache_control}]
            else:
                request_kwargs["system"] = [{"text": system_message, "type": "text"}]

        if tools:
            request_kwargs["tools"] = format_tools_for_model(tools)

        if request_kwargs:
            log_debug(f"Calling {self.provider} with request parameters: {request_kwargs}", log_level=2)
        return request_kwargs

    def invoke(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
    ) -> ModelResponse:
        """
        Send a request to the Anthropic API to generate a response.
        """
        try:
            if run_response and run_response.metrics:
                run_response.metrics.set_time_to_first_token()

            chat_messages, system_message = format_messages(messages)
            request_kwargs = self._prepare_request_kwargs(system_message, tools)

            if self.mcp_servers is not None:
                assistant_message.metrics.start_timer()
                provider_response = self.get_client().beta.messages.create(
                    model=self.id,
                    messages=chat_messages,  # type: ignore
                    **self.get_request_params(),
                )
            else:
                assistant_message.metrics.start_timer()
                provider_response = self.get_client().messages.create(
                    model=self.id,
                    messages=chat_messages,  # type: ignore
                    **request_kwargs,
                )

            assistant_message.metrics.stop_timer()

            # Parse the response into an Agno ModelResponse object
            model_response = self._parse_provider_response(provider_response, response_format=response_format)  # type: ignore

            return model_response

        except APIConnectionError as e:
            log_error(f"Connection error while calling Claude API: {str(e)}")
            raise ModelProviderError(message=e.message, model_name=self.name, model_id=self.id) from e
        except RateLimitError as e:
            log_warning(f"Rate limit exceeded: {str(e)}")
            raise ModelRateLimitError(message=e.message, model_name=self.name, model_id=self.id) from e
        except APIStatusError as e:
            log_error(f"Claude API error (status {e.status_code}): {str(e)}")
            raise ModelProviderError(
                message=e.message, status_code=e.status_code, model_name=self.name, model_id=self.id
            ) from e
        except Exception as e:
            log_error(f"Unexpected error calling Claude API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    def invoke_stream(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
    ) -> Any:
        """
        Stream a response from the Anthropic API.

        Args:
            messages (List[Message]): A list of messages to send to the model.

        Returns:
            Any: The streamed response from the model.

        Raises:
            APIConnectionError: If there are network connectivity issues
            RateLimitError: If the API rate limit is exceeded
            APIStatusError: For other API-related errors
        """
        chat_messages, system_message = format_messages(messages)
        request_kwargs = self._prepare_request_kwargs(system_message, tools)

        try:
            if run_response and run_response.metrics:
                run_response.metrics.set_time_to_first_token()

            if self.mcp_servers is not None:
                assistant_message.metrics.start_timer()
                with self.get_client().beta.messages.stream(
                    model=self.id,
                    messages=chat_messages,  # type: ignore
                    **request_kwargs,
                ) as stream:
                    for chunk in stream:
                        yield self._parse_provider_response_delta(chunk)  # type: ignore
            else:
                assistant_message.metrics.start_timer()
                with self.get_client().messages.stream(
                    model=self.id,
                    messages=chat_messages,  # type: ignore
                    **request_kwargs,
                ) as stream:
                    for chunk in stream:  # type: ignore
                        yield self._parse_provider_response_delta(chunk)  # type: ignore

            assistant_message.metrics.stop_timer()

        except APIConnectionError as e:
            log_error(f"Connection error while calling Claude API: {str(e)}")
            raise ModelProviderError(message=e.message, model_name=self.name, model_id=self.id) from e
        except RateLimitError as e:
            log_warning(f"Rate limit exceeded: {str(e)}")
            raise ModelRateLimitError(message=e.message, model_name=self.name, model_id=self.id) from e
        except APIStatusError as e:
            log_error(f"Claude API error (status {e.status_code}): {str(e)}")
            raise ModelProviderError(
                message=e.message, status_code=e.status_code, model_name=self.name, model_id=self.id
            ) from e
        except Exception as e:
            log_error(f"Unexpected error calling Claude API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    async def ainvoke(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
    ) -> ModelResponse:
        """
        Send an asynchronous request to the Anthropic API to generate a response.
        """
        try:
            if run_response and run_response.metrics:
                run_response.metrics.set_time_to_first_token()

            chat_messages, system_message = format_messages(messages)
            request_kwargs = self._prepare_request_kwargs(system_message, tools)

            if self.mcp_servers is not None:
                assistant_message.metrics.start_timer()
                provider_response = await self.get_async_client().beta.messages.create(
                    model=self.id,
                    messages=chat_messages,  # type: ignore
                    **self.get_request_params(),
                )
            else:
                assistant_message.metrics.start_timer()
                provider_response = await self.get_async_client().messages.create(
                    model=self.id,
                    messages=chat_messages,  # type: ignore
                    **request_kwargs,
                )

            assistant_message.metrics.stop_timer()

            # Parse the response into an Agno ModelResponse object
            model_response = self._parse_provider_response(provider_response, response_format=response_format)  # type: ignore

            return model_response

        except APIConnectionError as e:
            log_error(f"Connection error while calling Claude API: {str(e)}")
            raise ModelProviderError(message=e.message, model_name=self.name, model_id=self.id) from e
        except RateLimitError as e:
            log_warning(f"Rate limit exceeded: {str(e)}")
            raise ModelRateLimitError(message=e.message, model_name=self.name, model_id=self.id) from e
        except APIStatusError as e:
            log_error(f"Claude API error (status {e.status_code}): {str(e)}")
            raise ModelProviderError(
                message=e.message, status_code=e.status_code, model_name=self.name, model_id=self.id
            ) from e
        except Exception as e:
            log_error(f"Unexpected error calling Claude API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    async def ainvoke_stream(
        self,
        messages: List[Message],
        assistant_message: Message,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
    ) -> AsyncIterator[ModelResponse]:
        """
        Stream an asynchronous response from the Anthropic API.
        Args:
            messages (List[Message]): A list of messages to send to the model.
        Returns:
            AsyncIterator[ModelResponse]: An async iterator of processed model responses.
        Raises:
            APIConnectionError: If there are network connectivity issues
            RateLimitError: If the API rate limit is exceeded
            APIStatusError: For other API-related errors
        """
        try:
            if run_response and run_response.metrics:
                run_response.metrics.set_time_to_first_token()

            chat_messages, system_message = format_messages(messages)
            request_kwargs = self._prepare_request_kwargs(system_message, tools)

            if self.mcp_servers is not None:
                assistant_message.metrics.start_timer()
                async with self.get_async_client().beta.messages.stream(
                    model=self.id,
                    messages=chat_messages,  # type: ignore
                    **request_kwargs,
                ) as stream:
                    async for chunk in stream:
                        yield self._parse_provider_response_delta(chunk)  # type: ignore
            else:
                assistant_message.metrics.start_timer()
                async with self.get_async_client().messages.stream(
                    model=self.id,
                    messages=chat_messages,  # type: ignore
                    **request_kwargs,
                ) as stream:
                    async for chunk in stream:  # type: ignore
                        yield self._parse_provider_response_delta(chunk)  # type: ignore

            assistant_message.metrics.stop_timer()

        except APIConnectionError as e:
            log_error(f"Connection error while calling Claude API: {str(e)}")
            raise ModelProviderError(message=e.message, model_name=self.name, model_id=self.id) from e
        except RateLimitError as e:
            log_warning(f"Rate limit exceeded: {str(e)}")
            raise ModelRateLimitError(message=e.message, model_name=self.name, model_id=self.id) from e
        except APIStatusError as e:
            log_error(f"Claude API error (status {e.status_code}): {str(e)}")
            raise ModelProviderError(
                message=e.message, status_code=e.status_code, model_name=self.name, model_id=self.id
            ) from e
        except Exception as e:
            log_error(f"Unexpected error calling Claude API: {str(e)}")
            raise ModelProviderError(message=str(e), model_name=self.name, model_id=self.id) from e

    def get_system_message_for_model(self, tools: Optional[List[Any]] = None) -> Optional[str]:
        if tools is not None and len(tools) > 0:
            tool_call_prompt = "Do not reflect on the quality of the returned search results in your response\n\n"
            return tool_call_prompt
        return None

    def _parse_provider_response(self, response: AnthropicMessage, **kwargs) -> ModelResponse:
        """
        Parse the Claude response into a ModelResponse.

        Args:
            response: Raw response from Anthropic

        Returns:
            ModelResponse: Parsed response data
        """
        model_response = ModelResponse()

        # Add role (Claude always uses 'assistant')
        model_response.role = response.role or "assistant"

        if response.content:
            for block in response.content:
                if block.type == "text":
                    if model_response.content is None:
                        model_response.content = block.text
                    else:
                        model_response.content += block.text

                    # Capture citations from the response
                    if block.citations is not None:
                        if model_response.citations is None:
                            model_response.citations = Citations(raw=[], urls=[], documents=[])
                        for citation in block.citations:
                            model_response.citations.raw.append(citation.model_dump())  # type: ignore
                            # Web search citations
                            if isinstance(citation, CitationsWebSearchResultLocation):
                                model_response.citations.urls.append(  # type: ignore
                                    UrlCitation(url=citation.url, title=citation.cited_text)
                                )
                            # Document citations
                            elif isinstance(citation, CitationPageLocation):
                                model_response.citations.documents.append(  # type: ignore
                                    DocumentCitation(
                                        document_title=citation.document_title,
                                        cited_text=citation.cited_text,
                                    )
                                )
                elif block.type == "thinking":
                    model_response.reasoning_content = block.thinking
                    model_response.provider_data = {
                        "signature": block.signature,
                    }
                elif block.type == "redacted_thinking":
                    model_response.redacted_reasoning_content = block.data

        # Extract tool calls from the response
        if response.stop_reason == "tool_use":
            for block in response.content:
                if block.type == "tool_use":
                    tool_name = block.name
                    tool_input = block.input

                    function_def = {"name": tool_name}
                    if tool_input:
                        function_def["arguments"] = json.dumps(tool_input)

                    model_response.extra = model_response.extra or {}

                    model_response.tool_calls.append(
                        {
                            "id": block.id,
                            "type": "function",
                            "function": function_def,
                        }
                    )

        # Add usage metrics
        if response.usage is not None:
            model_response.response_usage = self._get_metrics(response.usage)

        return model_response

    def _parse_provider_response_delta(
        self,
        response: Union[
            ContentBlockStartEvent,
            ContentBlockDeltaEvent,
            ContentBlockStopEvent,
            MessageStopEvent,
            BetaRawContentBlockDeltaEvent,
        ],
    ) -> ModelResponse:
        """
        Parse the Claude streaming response into ModelProviderResponse objects.

        Args:
            response: Raw response chunk from Anthropic

        Returns:
            ModelResponse: Iterator of parsed response data
        """
        model_response = ModelResponse()

        if isinstance(response, ContentBlockStartEvent):
            if response.content_block.type == "redacted_reasoning_content":
                model_response.redacted_reasoning_content = response.content_block.data

        if isinstance(response, ContentBlockDeltaEvent):
            # Handle text content
            if response.delta.type == "text_delta":
                model_response.content = response.delta.text
            # Handle thinking content
            elif response.delta.type == "thinking_delta":
                model_response.reasoning_content = response.delta.thinking
            elif response.delta.type == "signature_delta":
                model_response.provider_data = {
                    "signature": response.delta.signature,
                }

        elif isinstance(response, ContentBlockStopEvent):
            if response.content_block.type == "tool_use":  # type: ignore
                tool_use = response.content_block  # type: ignore
                tool_name = tool_use.name
                tool_input = tool_use.input

                function_def = {"name": tool_name}
                if tool_input:
                    function_def["arguments"] = json.dumps(tool_input)

                model_response.extra = model_response.extra or {}

                model_response.tool_calls = [
                    {
                        "id": tool_use.id,
                        "type": "function",
                        "function": function_def,
                    }
                ]

        # Capture citations from the final response
        elif isinstance(response, MessageStopEvent):
            model_response.content = ""
            model_response.citations = Citations(raw=[], urls=[], documents=[])
            for block in response.message.content:  # type: ignore
                citations = getattr(block, "citations", None)
                if not citations:
                    continue
                for citation in citations:
                    model_response.citations.raw.append(citation.model_dump())  # type: ignore
                    # Web search citations
                    if isinstance(citation, CitationsWebSearchResultLocation):
                        model_response.citations.urls.append(UrlCitation(url=citation.url, title=citation.cited_text))  # type: ignore
                    # Document citations
                    elif isinstance(citation, CitationPageLocation):
                        model_response.citations.documents.append(  # type: ignore
                            DocumentCitation(document_title=citation.document_title, cited_text=citation.cited_text)
                        )

        if hasattr(response, "message") and hasattr(response.message, "usage") and response.message.usage is not None:  # type: ignore
            model_response.response_usage = self._get_metrics(response.message.usage)  # type: ignore

        # Capture the Beta response
        try:
            if (
                isinstance(response, BetaRawContentBlockDeltaEvent)
                and isinstance(response.delta, BetaTextDelta)
                and response.delta.text is not None
            ):
                model_response.content = response.delta.text
        except Exception as e:
            log_error(f"Error parsing Beta response: {e}")

        return model_response

    def _get_metrics(self, response_usage: Union[Usage, MessageDeltaUsage]) -> Metrics:
        """
        Parse the given Anthropic-specific usage into an Agno Metrics object.

        Args:
            response_usage: Usage data from Anthropic

        Returns:
            Metrics: Parsed metrics data
        """
        metrics = Metrics()

        metrics.input_tokens = response_usage.input_tokens or 0
        metrics.output_tokens = response_usage.output_tokens or 0
        metrics.total_tokens = metrics.input_tokens + metrics.output_tokens
        metrics.cache_read_tokens = response_usage.cache_read_input_tokens or 0
        metrics.cache_write_tokens = response_usage.cache_creation_input_tokens or 0

        # Anthropic-specific additional fields
        if response_usage.server_tool_use:
            metrics.provider_metrics = {"server_tool_use": response_usage.server_tool_use}
        if isinstance(response_usage, Usage):
            if response_usage.service_tier:
                metrics.provider_metrics = metrics.provider_metrics or {}
                metrics.provider_metrics["service_tier"] = response_usage.service_tier

        return metrics



from dataclasses import dataclass, field
from os import getenv
from typing import Any, Dict, List, Optional, Type, Union

from pydantic import BaseModel

from agno.models.openai.like import OpenAILike
from agno.run.agent import RunOutput


@dataclass
class OpenRouter(OpenAILike):
    """
    A class for using models hosted on OpenRouter.

    Attributes:
        id (str): The model id. Defaults to "gpt-4o".
        name (str): The model name. Defaults to "OpenRouter".
        provider (str): The provider name. Defaults to "OpenRouter".
        api_key (Optional[str]): The API key.
        base_url (str): The base URL. Defaults to "https://openrouter.ai/api/v1".
        max_tokens (int): The maximum number of tokens. Defaults to 1024.
        fallback_models (Optional[List[str]]): List of fallback model IDs to use if the primary model
            fails due to rate limits, timeouts, or unavailability. OpenRouter will automatically try
            these models in order. Example: ["anthropic/claude-sonnet-4", "deepseek/deepseek-r1"]
    """

    id: str = "gpt-4o"
    name: str = "OpenRouter"
    provider: str = "OpenRouter"

    api_key: Optional[str] = field(default_factory=lambda: getenv("OPENROUTER_API_KEY"))
    base_url: str = "https://openrouter.ai/api/v1"
    max_tokens: int = 1024
    models: Optional[List[str]] = None  # Dynamic model routing https://openrouter.ai/docs/features/model-routing

    def get_request_params(
        self,
        response_format: Optional[Union[Dict, Type[BaseModel]]] = None,
        tools: Optional[List[Dict[str, Any]]] = None,
        tool_choice: Optional[Union[str, Dict[str, Any]]] = None,
        run_response: Optional[RunOutput] = None,
    ) -> Dict[str, Any]:
        """
        Returns keyword arguments for API requests, including fallback models configuration.

        Returns:
            Dict[str, Any]: A dictionary of keyword arguments for API requests.
        """
        # Get base request params from parent class
        request_params = super().get_request_params(
            response_format=response_format, tools=tools, tool_choice=tool_choice, run_response=run_response
        )

        # Add fallback models to extra_body if specified
        if self.models:
            # Get existing extra_body or create new dict
            extra_body = request_params.get("extra_body") or {}

            # Merge fallback models into extra_body
            extra_body["models"] = self.models

            # Update request params
            request_params["extra_body"] = extra_body

        return request_params



from datetime import datetime
from typing import Any, Dict, Optional

from pydantic import BaseModel, ConfigDict, model_validator


class KnowledgeRow(BaseModel):
    """Knowledge Row that is stored in the database"""

    # id for this knowledge, auto-generated if not provided
    id: Optional[str] = None
    name: str
    description: str
    metadata: Optional[Dict[str, Any]] = None
    type: Optional[str] = None
    size: Optional[int] = None
    linked_to: Optional[str] = None
    access_count: Optional[int] = None
    status: Optional[str] = None
    status_message: Optional[str] = None
    created_at: Optional[int] = None
    updated_at: Optional[int] = None
    external_id: Optional[str] = None

    model_config = ConfigDict(from_attributes=True, arbitrary_types_allowed=True)

    @model_validator(mode="after")
    def generate_id(self) -> "KnowledgeRow":
        if self.id is None:
            from uuid import uuid4

            self.id = str(uuid4())
        return self

    def to_dict(self) -> Dict[str, Any]:
        _dict = self.model_dump(exclude={"updated_at"})

        _dict["updated_at"] = datetime.fromtimestamp(self.updated_at).isoformat() if self.updated_at else None

        return _dict
